from __future__ import absolute_import

from .ABuPickRegressAngMinMax import AbuPickRegressAngMinMax
from .ABuPickSimilarNTop import AbuPickSimilarNTop
from .ABuPickStockBase import AbuPickStockBase
from .ABuPickStockPriceMinMax import AbuPickStockPriceMinMax
from .ABuPickStockDemo import AbuPickStockShiftDistance, AbuPickStockNTop
from . import ABuPickStock as ps

__all__ = [
    'AbuPickRegressAngMinMax',
    'AbuPickSimilarNTop',
    'AbuPickStockBase',
    'AbuPickStockPriceMinMax',
    'AbuPickStockShiftDistance',
    'AbuPickStockNTop',
    'ps']



# -*- encoding:utf-8 -*-
"""
    深度学习工具模块，为caffe工具库做数据集准备，
    切割训练集
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from collections import defaultdict
import itertools
import os

__author__ = '阿布'
__weixin__ = 'abu_quant'

__all__ = ['train_val_split']


def train_val_split(train_path, n_folds=10):
    """将caffe返回的数据集文件进行切割工作，切割为训练集，交叉测试集，测试集数据文件"""

    if n_folds <= 1:
        raise ValueError('n_folds must > 1')

    with open(train_path, 'r') as f:
        lines = f.readlines()
        class_dict = defaultdict(list)
        for line in lines:
            cs = line[line.rfind(' '):]
            class_dict[cs].append(line)

    train = list()
    val = list()
    for cs in class_dict:
        cs_len = len(class_dict[cs])
        val_cnt = int(cs_len / n_folds)
        val.append(class_dict[cs][:val_cnt])
        train.append(class_dict[cs][val_cnt:])
    val = list(itertools.chain.from_iterable(val))
    train = list(itertools.chain.from_iterable(train))
    test = [t.split(' ')[0] + '\n' for t in val]

    # 在参数的train_path同目录下写train_split.txt
    fn = os.path.dirname(train_path) + '/train_split.txt'
    with open(fn, 'wb') as f:
        f.writelines(train)
    # 在参数的train_path同目录下写val_split.txt
    fn = os.path.dirname(train_path) + '/val_split.txt'
    with open(fn, 'wb') as f:
        f.writelines(val)
    # 在参数的train_path同目录下写/test_split.txt
    fn = os.path.dirname(train_path) + '/test_split.txt'
    with open(fn, 'wb') as f:
        f.writelines(test)



from __future__ import absolute_import

from . import ABuTLine as line
from . import ABuTLExecute as execute
from . import ABuTLAtr as atr
from . import ABuTLGolden as golden
from . import ABuTLJump as jump
from . import ABuTLSimilar as similar
from . import ABuTLVwap as vwap
from . import ABuTLWave as wave
from .ABuTLine import ESkeletonHow, EShiftDistanceHow, AbuTLine

__all__ = [
    'line',
    'execute',
    'atr',
    'golden',
    'jump',
    'similar',
    'vwap',
    'wave',

    'ESkeletonHow',
    'EShiftDistanceHow',
    'AbuTLine']



# -*- encoding:utf-8 -*-
"""
    选股示例因子：价格选股因子
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from .ABuPickStockBase import AbuPickStockBase, reversed_result
import numpy as np

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuPickStockPriceMinMax(AbuPickStockBase):
    """价格选股因子示例类"""
    def _init_self(self, **kwargs):
        """通过kwargs设置选股价格边际条件，配置因子参数"""

        # 暂时与base保持一致不使用kwargs.pop('a', default)方式
        # fit_pick中选择 > 最小(threshold_price_min), 默认负无穷，即默认所有都符合
        self.threshold_price_min = -np.inf
        if 'threshold_price_min' in kwargs:
            # 最小价格阀值
            self.threshold_price_min = kwargs['threshold_price_min']

        # fit_pick中选择 < 最大(threshold_price_max), 默认正无穷，即默认所有都符合
        self.threshold_price_max = np.inf
        if 'threshold_price_max' in kwargs:
            # 最大价格阀值
            self.threshold_price_max = kwargs['threshold_price_max']

    @reversed_result
    def fit_pick(self, kl_pd, target_symbol):
        """开始根据自定义价格边际参数进行选股"""
        if kl_pd.close.max() < self.threshold_price_max and kl_pd.close.min() > self.threshold_price_min:
            # kl_pd.close的最大价格 < 最大价格阀值 且 kl_pd.close的最小价格 > 最小价格阀值
            return True
        return False

    def fit_first_choice(self, pick_worker, choice_symbols, *args, **kwargs):
        raise NotImplementedError('AbuPickStockPriceMinMax fit_first_choice unsupported now!')



# -*- encoding:utf-8 -*-
"""度量模块基础"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from ..ExtBu.empyrical import stats
from ..CoreBu import ABuEnv
from ..CoreBu.ABuEnv import EMarketDataFetchMode
from ..CoreBu.ABuFixes import six
from ..UtilBu import ABuDateUtil
from ..UtilBu import ABuStatsUtil, ABuScalerUtil
from ..UtilBu.ABuDTUtil import warnings_filter
from ..TradeBu.ABuKLManager import AbuKLManager
from ..TradeBu.ABuCapital import AbuCapital
from ..TradeBu import ABuTradeExecute


__author__ = '阿布'
__weixin__ = 'abu_quant'


def valid_check(func):
    """检测度量的输入是否正常，非正常显示info，正常继续执行被装饰方法"""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.valid:
            return func(self, *args, **kwargs)
        else:
            logging.info('metrics input is invalid or zero order gen!')

    return wrapper


class AbuMetricsBase(object):
    """主要适配股票类型交易对象的回测结果度量"""

    @classmethod
    def show_general(cls, orders_pd, action_pd, capital, benchmark, returns_cmp=False,
                     only_info=False, only_show_returns=False, enable_stocks_full_rate_factor=False):
        """
        类方法，针对输入执行度量后执行主要度量可视化及度量结果信息输出
        :param orders_pd: 回测结果生成的交易订单构成的pd.DataFrame对象
        :param action_pd: 回测结果生成的交易行为构成的pd.DataFrame对象
        :param capital: 资金类AbuCapital实例化对象
        :param benchmark: 交易基准对象，AbuBenchmark实例对象
        :param returns_cmp: 是否只度量无资金管理的情况下总体情况
        :param only_info: 是否只显示文字度量结果，不显示图像
        :param only_show_returns: 透传plot_returns_cmp，默认False, True则只显示收益对比不显示其它可视化
        :param enable_stocks_full_rate_factor: 是否开启满仓乘数
        :return AbuMetricsBase实例化类型对象
        """
        metrics = cls(orders_pd, action_pd, capital, benchmark,
                      enable_stocks_full_rate_factor=enable_stocks_full_rate_factor)
        metrics.fit_metrics()
        if returns_cmp:
            metrics.plot_order_returns_cmp(only_info=only_info)
        else:
            metrics.plot_returns_cmp(only_info=only_info, only_show_returns=only_show_returns)
            if not only_show_returns:
                metrics.plot_sharp_volatility_cmp(only_info=only_info)
        return metrics

    def __init__(self, orders_pd, action_pd, capital, benchmark, enable_stocks_full_rate_factor=False):
        """
        :param orders_pd: 回测结果生成的交易订单构成的pd.DataFrame对象
        :param action_pd: 回测结果生成的交易行为构成的pd.DataFrame对象
        :param capital: 资金类AbuCapital实例化对象
        :param benchmark: 交易基准对象，AbuBenchmark实例对象
        :param enable_stocks_full_rate_factor: 是否开启满仓乘数
        """
        self.capital = capital
        self.orders_pd = orders_pd
        self.action_pd = action_pd
        self.benchmark = benchmark
        """
            满仓乘数，如果设置为True, 针对度量信息如收益等需要除self.stocks_full_rate_factor
        """
        self.enable_stocks_full_rate_factor = enable_stocks_full_rate_factor
        # 验证输入的回测数据是否可度量，便于valid_check装饰器工作
        self.valid = False
        if self.capital is not None and 'capital_blance' in self.capital.capital_pd:
            self.valid = True
        # ipython notebook下使用logging.info
        self.log_func = logging.info if ABuEnv.g_is_ipython else print

    @valid_check
    def fit_metrics(self):
        """执行所有度量函数"""
        # TODO 根据ORDER数量大于一定阀值启动进度条
        # with AbuProgress(100, 0, label='metrics progress...') as pg:
        # pg.show(5)
        self._metrics_base_stats()
        # pg.show(50)
        self._metrics_sell_stats()
        # pg.show(80)
        self._metrics_action_stats()
        # pg.show(95)
        self._metrics_extend_stats()

    def fit_metrics_order(self):
        """对外接口，并非度量真实成交了的结果，只度量orders_pd，即不涉及资金的度量"""
        self._metrics_sell_stats()

    def _metrics_base_stats(self):
        """度量真实成交了的capital_pd，即涉及资金的度量"""
        # 平均资金利用率
        self.cash_utilization = 1 - (self.capital.capital_pd.cash_blance /
                                     self.capital.capital_pd.capital_blance).mean()

        # 默认不使用满仓乘数即stocks_full_rate_factor＝1
        self.stocks_full_rate_factor = 1
        if self.enable_stocks_full_rate_factor:
            # 计算满仓比例
            stocks_full_rate = (self.capital.capital_pd.stocks_blance / self.capital.capital_pd.capital_blance)
            # 避免除0
            stocks_full_rate[stocks_full_rate == 0] = 1
            # 倒数得到满仓乘数
            self.stocks_full_rate_factor = (1 / stocks_full_rate)

        # 收益数据
        self.benchmark_returns = np.round(self.benchmark.kl_pd.close.pct_change(), 3)
        # 如果enable_stocks_full_rate_factor 则 * self.stocks_full_rate_factor的意义为随时都是满仓
        self.algorithm_returns = np.round(self.capital.capital_pd['capital_blance'].pct_change(),
                                          3) * self.stocks_full_rate_factor

        # 收益cum数据
        # noinspection PyTypeChecker
        self.algorithm_cum_returns = stats.cum_returns(self.algorithm_returns)
        self.benchmark_cum_returns = stats.cum_returns(self.benchmark_returns)

        # 最后一日的cum return
        self.benchmark_period_returns = self.benchmark_cum_returns[-1]
        self.algorithm_period_returns = self.algorithm_cum_returns[-1]

        # 交易天数
        self.num_trading_days = len(self.benchmark_returns)

        # 年化收益
        self.algorithm_annualized_returns = \
            (ABuEnv.g_market_trade_year / self.num_trading_days) * self.algorithm_period_returns
        self.benchmark_annualized_returns = \
            (ABuEnv.g_market_trade_year / self.num_trading_days) * self.benchmark_period_returns

        # 策略平均收益
        # noinspection PyUnresolvedReferences
        self.mean_algorithm_returns = self.algorithm_returns.cumsum() / np.arange(1, self.num_trading_days + 1,
                                                                                  dtype=np.float64)
        # 波动率
        self.benchmark_volatility = stats.annual_volatility(self.benchmark_returns)
        # noinspection PyTypeChecker
        self.algorithm_volatility = stats.annual_volatility(self.algorithm_returns)

        # 夏普比率
        self.benchmark_sharpe = stats.sharpe_ratio(self.benchmark_returns)
        # noinspection PyTypeChecker
        self.algorithm_sharpe = stats.sharpe_ratio(self.algorithm_returns)

        # 信息比率
        # noinspection PyUnresolvedReferences
        self.information = stats.information_ratio(self.algorithm_returns.values, self.benchmark_returns.values)

        # 阿尔法, 贝塔
        # noinspection PyUnresolvedReferences
        self.alpha, self.beta = stats.alpha_beta_aligned(self.algorithm_returns.values, self.benchmark_returns.values)

        # 最大回撤
        # noinspection PyUnresolvedReferences
        self.max_drawdown = stats.max_drawdown(self.algorithm_returns.values)

    def _metrics_sell_stats(self):
        """并非度量真实成交了的结果，只度量orders_pd，即认为没有仓位管理和资金量限制前提下的表现"""

        # 根据order中的数据，计算盈利比例
        self.orders_pd['profit_cg'] = self.orders_pd['profit'] / (
            self.orders_pd['buy_price'] * self.orders_pd['buy_cnt'])
        # 为了显示方便及明显
        self.orders_pd['profit_cg_hunder'] = self.orders_pd['profit_cg'] * 100
        # 成交了的pd isin win or loss
        deal_pd = self.orders_pd[self.orders_pd['sell_type'].isin(['win', 'loss'])]
        # 卖出原因get_dummies进行离散化
        dumm_sell = pd.get_dummies(deal_pd.sell_type_extra)
        dumm_sell_t = dumm_sell.T
        # 为plot_sell_factors函数生成卖出生效因子分布
        self.dumm_sell_t_sum = dumm_sell_t.sum(axis=1)

        # 买入因子唯一名称get_dummies进行离散化
        dumm_buy = pd.get_dummies(deal_pd.buy_factor)
        dumm_buy = dumm_buy.T
        # 为plot_buy_factors函数生成卖出生效因子分布
        self.dumm_buy_t_sum = dumm_buy.sum(axis=1)

        self.orders_pd['buy_date'] = self.orders_pd['buy_date'].astype(int)
        self.orders_pd[self.orders_pd['result'] != 0]['sell_date'].astype(int, copy=False)
        # 因子的单子的持股时间长度计算
        self.orders_pd['keep_days'] = self.orders_pd.apply(lambda x:
                                                           ABuDateUtil.diff(x['buy_date'],
                                                                            ABuDateUtil.current_date_int()
                                                                            if x['result'] == 0 else x[
                                                                                'sell_date']),
                                                           axis=1)
        # 筛出已经成交了的单子
        self.order_has_ret = self.orders_pd[self.orders_pd['result'] != 0]

        # 筛出未成交的单子
        self.order_keep = self.orders_pd[self.orders_pd['result'] == 0]

        xt = self.order_has_ret.result.value_counts()
        # 计算胜率
        if xt.shape[0] == 2:
            win_rate = xt[1] / xt.sum()
        elif xt.shape[0] == 1:
            win_rate = xt.index[0]
        else:
            win_rate = 0
        self.win_rate = win_rate
        # 策略持股天数平均值
        self.keep_days_mean = self.orders_pd['keep_days'].mean()
        # 策略持股天数中位数
        self.keep_days_median = self.orders_pd['keep_days'].median()

        # 策略期望收益
        self.gains_mean = self.order_has_ret[self.order_has_ret['profit_cg'] > 0].profit_cg.mean()
        if np.isnan(self.gains_mean):
            self.gains_mean = 0.0
        # 策略期望亏损
        self.losses_mean = self.order_has_ret[self.order_has_ret['profit_cg'] < 0].profit_cg.mean()
        if np.isnan(self.losses_mean):
            self.losses_mean = 0.0

        # 忽略仓位控的前提下，即假设每一笔交易使用相同的资金，策略的总获利交易获利比例和
        profit_cg_win_sum = self.order_has_ret[self.order_has_ret['profit_cg'] > 0].profit.sum()
        # 忽略仓位控的前提下，即假设每一笔交易使用相同的资金，策略的总亏损交易亏损比例和
        profit_cg_loss_sum = self.order_has_ret[self.order_has_ret['profit_cg'] < 0].profit.sum()

        if profit_cg_win_sum * profit_cg_loss_sum == 0 and profit_cg_win_sum + profit_cg_loss_sum > 0:
            # 其中有一个是0的，要转换成一个最小统计单位计算盈亏比，否则不需要
            if profit_cg_win_sum == 0:
                profit_cg_win_sum = 0.01
            if profit_cg_loss_sum == 0:
                profit_cg_win_sum = 0.01

        #  忽略仓位控的前提下，计算盈亏比
        self.win_loss_profit_rate = 0 if profit_cg_loss_sum == 0 else -round(profit_cg_win_sum / profit_cg_loss_sum, 4)
        #  忽略仓位控的前提下，计算所有交易单的盈亏总会
        self.all_profit = self.order_has_ret['profit'].sum()

    def _metrics_action_stats(self):
        """度量真实成交了的action_pd 计算买入资金的分布平均性，及是否有良好的分布"""

        action_pd = self.action_pd
        # 只选生效的, 由于忽略非交易日, 大概有多出0.6的误差
        self.act_buy = action_pd[action_pd.action.isin(['buy']) & action_pd.deal.isin([True])]
        # drop重复的日期上的行为，只保留一个，cp_date形如下所示
        cp_date = self.act_buy['Date'].drop_duplicates()
        """
            cp_date
            0      20141024
            2      20141029
            20     20150127
            21     20150205
            23     20150213
            25     20150218
            31     20150310
            34     20150401
            36     20150409
            39     20150422
            41     20150423
            44     20150428
            58     20150609
            59     20150610
            63     20150624
            66     20150715
            67     20150717
        """
        dt_fmt = cp_date.apply(lambda order: ABuDateUtil.str_to_datetime(str(order), '%Y%m%d'))
        dt_fmt = dt_fmt.apply(lambda order: (order - dt_fmt.iloc[0]).days)
        # 前后两两生效交易时间相减
        self.diff_dt = dt_fmt - dt_fmt.shift(1)
        # 计算平均生效间隔时间
        self.effect_mean_day = self.diff_dt.mean()

        self.act_buy['cost'] = self.act_buy.apply(lambda order: order.Price * order.Cnt, axis=1)
        # 计算cost各种统计度量值
        self.cost_stats = ABuStatsUtil.stats_namedtuple(self.act_buy['cost'])

        buy_action_pd = action_pd[action_pd['action'] == 'buy']
        buy_action_pd_deal = buy_action_pd['deal']
        # 计算资金对应的成交比例
        self.buy_deal_rate = buy_action_pd_deal.sum() / buy_action_pd_deal.count()

    def _metrics_extend_stats(self):
        """子类可扩展的metrics方法，子类在此方法中可定义自己需要度量的值"""
        pass

    @valid_check
    @warnings_filter  # skip: statsmodels / nonparametric / kdetools.py:20
    def plot_order_returns_cmp(self, only_info=True):
        """非真实成交的度量，认为资金无限，无资金管理的情况下总体情况"""

        self.log_func('买入后卖出的交易数量:{}'.format(self.order_has_ret.shape[0]))
        self.log_func('买入后尚未卖出的交易数量:{}'.format(self.order_keep.shape[0]))
        self.log_func('胜率:{:.4f}%'.format(self.win_rate * 100))
        self.log_func('平均获利期望:{:.4f}%'.format(self.gains_mean * 100))
        self.log_func('平均亏损期望:{:.4f}%'.format(self.losses_mean * 100))
        self.log_func('盈亏比:{:.4f}'.format(self.win_loss_profit_rate))
        self.log_func('所有交易收益比例和:{:.4f} '.format(self.order_has_ret.profit_cg.sum()))
        self.log_func('所有交易总盈亏和:{:.4f} '.format(self.all_profit))

        if only_info:
            return
        # 无法与基准对比，只能表示取向
        self.order_has_ret.sort_values('buy_date')['profit_cg'].cumsum().plot(grid=True, title='profit_cg cumsum')
        plt.show()

    @valid_check
    def plot_returns_cmp(self, only_show_returns=False, only_info=False):
        """考虑资金情况下的度量，进行与benchmark的收益度量对比，收益趋势，资金变动可视化，以及其它度量信息"""

        self.log_func('买入后卖出的交易数量:{}'.format(self.order_has_ret.shape[0]))
        self.log_func('买入后尚未卖出的交易数量:{}'.format(self.order_keep.shape[0]))

        self.log_func('胜率:{:.4f}%'.format(self.win_rate * 100))

        self.log_func('平均获利期望:{:.4f}%'.format(self.gains_mean * 100))
        self.log_func('平均亏损期望:{:.4f}%'.format(self.losses_mean * 100))

        self.log_func('盈亏比:{:.4f}'.format(self.win_loss_profit_rate))

        self.log_func('策略收益: {:.4f}%'.format(self.algorithm_period_returns * 100))
        self.log_func('基准收益: {:.4f}%'.format(self.benchmark_period_returns * 100))
        self.log_func('策略年化收益: {:.4f}%'.format(self.algorithm_annualized_returns * 100))
        self.log_func('基准年化收益: {:.4f}%'.format(self.benchmark_annualized_returns * 100))

        self.log_func('策略买入成交比例:{:.4f}%'.format(self.buy_deal_rate * 100))
        self.log_func('策略资金利用率比例:{:.4f}%'.format(self.cash_utilization * 100))
        self.log_func('策略共执行{}个交易日'.format(self.num_trading_days))

        if only_info:
            return

        self.benchmark_cum_returns.plot()
        self.algorithm_cum_returns.plot()
        plt.legend(['benchmark returns', 'algorithm returns'], loc='best')
        plt.show()

        if only_show_returns:
            return
        sns.regplot(x=np.arange(0, len(self.algorithm_cum_returns)), y=self.algorithm_cum_returns.values)
        plt.show()
        sns.distplot(self.capital.capital_pd['capital_blance'], kde_kws={"lw": 3, "label": "capital blance kde"})
        plt.show()

    @valid_check
    def plot_sharp_volatility_cmp(self, only_info=False):
        """sharp，volatility的策略与基准对比可视化，以及alpha阿尔法，beta贝塔，Information信息比率等信息输出"""

        self.log_func('alpha阿尔法:{:.4f}'.format(self.alpha))
        self.log_func('beta贝塔:{:.4f}'.format(self.beta))
        self.log_func('Information信息比率:{:.4f}'.format(self.information))

        self.log_func('策略Sharpe夏普比率: {:.4f}'.format(self.algorithm_sharpe))
        self.log_func('基准Sharpe夏普比率: {:.4f}'.format(self.benchmark_sharpe))

        self.log_func('策略波动率Volatility: {:.4f}'.format(self.algorithm_volatility))
        self.log_func('基准波动率Volatility: {:.4f}'.format(self.benchmark_volatility))

        if only_info:
            return

        sharp_volatility = pd.DataFrame([[self.algorithm_sharpe, self.benchmark_sharpe],
                                         [self.algorithm_volatility, self.benchmark_volatility]])
        sharp_volatility.columns = ['algorithm', 'benchmark']
        sharp_volatility.index = ['sharpe', 'volatility']
        sharp_volatility.plot(kind='bar', alpha=0.5)
        _ = plt.setp(plt.gca().get_xticklabels(), rotation=30)

    @valid_check
    def plot_effect_mean_day(self):
        """可视化因子平均生效间隔时间"""

        self.log_func('因子平均生效间隔:{}'.format(self.effect_mean_day))

        ddvc = self.diff_dt.value_counts()
        ddvc_rt = ddvc / ddvc.sum()
        plt.figure(figsize=(6, 6))
        plt.axes([0.025, 0.025, 0.95, 0.95])
        x = ddvc_rt.values
        labels = ddvc_rt.index
        plt.pie(x, labels=labels, explode=x * 0.1)
        plt.title('factor diff effect day')
        plt.show()

    @valid_check
    def plot_action_buy_cost(self):
        """可视化开仓花费情况"""

        self.log_func('开仓花费情况: ')
        self.log_func(self.cost_stats)

        plt.title('action buy cost')
        bins = int(len(self.act_buy['cost']) / 10)
        bins = bins if bins > 0 else 10
        self.act_buy['cost'].plot(kind='hist', bins=bins)
        plt.show()

    @valid_check
    def plot_sell_factors(self):
        """可视化卖出生效因子分布"""
        self.log_func('卖出择时生效因子分布：')
        self.log_func(self.dumm_sell_t_sum)
        if self.dumm_sell_t_sum.shape[0] > 1:
            self.dumm_sell_t_sum.plot(kind='barh')
            plt.title('sell factors barh')
            plt.show()

    @valid_check
    def plot_buy_factors(self):
        """可视化买入生效因子分布"""
        self.log_func('买入择时生效因子分布：')
        self.log_func(self.dumm_buy_t_sum)

        if self.dumm_buy_t_sum.shape[0] > 1:
            self.dumm_buy_t_sum.plot(kind='barh')
            plt.title('buy factors barh')
            plt.show()

    @valid_check
    def plot_keep_days(self):
        """可视化策略持股天数"""

        self.log_func('策略持股天数平均数: {:.3f}'.format(self.keep_days_mean))
        self.log_func('策略持股天数中位数: {:.3f}'.format(self.keep_days_median))
        bins = int(self.orders_pd['keep_days'].shape[0] / 5)
        bins = bins if bins > 0 else 5
        self.orders_pd['keep_days'].plot(kind='hist', bins=bins)
        plt.show()

    @valid_check
    def plot_max_draw_down(self):
        """可视化最大回撤"""

        cb_earn = self.capital.capital_pd['capital_blance'] - self.capital.read_cash
        shift = cb_earn.shape[0]
        max_draw_down = {-1: -1}
        cap_pd_index = cb_earn.index.tolist()

        for sf in np.arange(1, shift):
            sub_val = cb_earn.iloc[sf]
            sf_val = cb_earn[:sf]
            sf_val = sf_val.drop_duplicates(keep='last')

            diff = sf_val.values - sub_val

            if diff.max() > list(six.itervalues(max_draw_down))[0]:
                st_ind = diff.argmax()
                st_ind = sf_val.index[st_ind]
                end_ind = cap_pd_index[sf]
                max_draw_down = {(st_ind, end_ind): diff.max()}

        down_rate = list(six.itervalues(max_draw_down))[0] / self.capital.capital_pd['capital_blance'].loc[
            list(six.iterkeys(max_draw_down))[0][0]]
        """
            截取开始交易部分
        """
        cb_earn = cb_earn.loc[cb_earn[cb_earn != 0].index[0]:]
        cb_earn.plot()
        plt.plot(list(six.iterkeys(max_draw_down))[0][0], cb_earn.loc[list(six.iterkeys(max_draw_down))[0][0]],
                 'ro', markersize=12,
                 markeredgewidth=1.5,
                 markerfacecolor='None', markeredgecolor='green')

        plt.plot(list(six.iterkeys(max_draw_down))[0][1], cb_earn.loc[list(six.iterkeys(max_draw_down))[0][1]],
                 'ro', markersize=12,
                 markeredgewidth=1.5,
                 markerfacecolor='None', markeredgecolor='red')

        plt.plot([list(six.iterkeys(max_draw_down))[0][0], list(six.iterkeys(max_draw_down))[0][1]],
                 [cb_earn.loc[list(six.iterkeys(max_draw_down))[0][0]],
                  cb_earn.loc[list(six.iterkeys(max_draw_down))[0][1]]], 'o-')
        plt.grid(True)
        plt.show()

        self.log_func('最大回撤: {:5f}'.format(down_rate))
        self.log_func('最大回测启始时间:{}, 结束时间{}, 共回测{:3f}'.format(
            ABuDateUtil.timestamp_to_str(list(six.iterkeys(max_draw_down))[0][0]),
            ABuDateUtil.timestamp_to_str(list(six.iterkeys(max_draw_down))[0][1]),
            list(six.itervalues(max_draw_down))[0]))

    @valid_check
    def transform_to_full_rate_factor(self, read_cash=-1, kl_pd_manager=None, n_process_kl=ABuEnv.g_cpu_cnt,
                                      show=True):
        if ABuEnv.g_data_fetch_mode != EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL:
            self.log_func('transform_to_full_rate_factor func must in E_DATA_FETCH_FORCE_LOCAL env!')
            return

        if not hasattr(self, 'full_rate_metrics'):
            if read_cash == -1:
                # 如果外部不设置资金数，设置一个亿为大资金数
                read_cash = 100000000

            target_symbols = list(set(self.orders_pd.symbol))
            # 重新以很大的资金初始化AbuCapital
            capital = AbuCapital(read_cash, self.benchmark,
                                 user_commission_dict=self.capital.commission.commission_dict)
            if kl_pd_manager is None:
                kl_pd_manager = AbuKLManager(self.benchmark, capital)
                # 一次性在主进程中执行多进程获取k线数据，全部放入kl_pd_manager中，内部启动n_process_kl个进程执行
                kl_pd_manager.batch_get_pick_time_kl_pd(target_symbols, n_process=n_process_kl)

            # noinspection PyUnresolvedReferences
            action_pd = self.action_pd.sort_values(['Date', 'action'])
            action_pd.index = np.arange(0, action_pd.shape[0])
            # 最后将所有的action作用在资金上，生成资金时序，及判断是否能买入
            ABuTradeExecute.apply_action_to_capital(capital, action_pd, kl_pd_manager)
            # 最终创建一个子AbuMetricsBase对象在内部，action_pd, capital使用新计算出来的，满仓乘数参数设置为True
            # noinspection PyAttributeOutsideInit
            self.full_rate_metrics = AbuMetricsBase(self.orders_pd, action_pd, capital, self.benchmark,
                                                    enable_stocks_full_rate_factor=True)
            self.full_rate_metrics.fit_metrics()
        if show:
            self.full_rate_metrics.plot_returns_cmp(only_show_returns=True)
        return self.full_rate_metrics


class MetricsDemo(AbuMetricsBase):
    """
        扩展自定义度量类示例

        eg:
            metrics = MetricsDemo(*abu_result_tuple)
            metrics.fit_metrics()
            metrics.plot_commission()
    """

    def _metrics_extend_stats(self):
        """
            子类可扩展的metrics方法，子类在此方法中可定义自己需要度量的值:
            本demo示例交易手续费和策略收益之间的度量对比
        """
        commission_df = self.capital.commission.commission_df
        commission_df['commission'] = commission_df.commission.astype(float)
        commission_df['cumsum'] = commission_df.commission.cumsum()
        """
            eg:
                type	date	symbol	commission	cumsum
            0	buy	20141024	usAAPL	19.04	19.04
            0	buy	20141024	usAAPL	19.04	38.08
            0	buy	20141029	usNOAH	92.17	130.25
            0	buy	20141029	usBIDU	7.81	138.06
            0	buy	20141029	usBIDU	7.81	145.87
            0	buy	20141029	usVIPS	60.95	206.82
        """
        # 讲date转换为index
        dates_pd = pd.to_datetime(commission_df.date)
        commission = pd.DataFrame(index=dates_pd)
        """
            eg: commission
            2014-10-24	19.04
            2014-10-24	38.08
            2014-10-29	130.25
            2014-10-29	138.06
            2014-10-29	145.87
            2014-10-29	206.82
            2014-11-03	265.82
            2014-11-11	360.73
        """
        commission['cum'] = commission_df['cumsum'].values
        self.commission_cum = commission['cum']
        self.commission_sum = self.commission_cum[-1]

    def plot_commission(self):
        """
            使用计算好的首先费cumsum序列和策略收益cumsum序列进行可视化对比
            可视化收益曲线和手续费曲线之前的关系
        """
        print('回测周期内手续费共: {:.2f}'.format(self.commission_sum))
        # 使用缩放scaler_xy将两条曲线缩放到同一个级别
        x, y = ABuScalerUtil.scaler_xy(self.commission_cum, self.algorithm_cum_returns, type_look='look_max',
                                       mean_how=True)
        x.plot(label='commission')
        y.plot(label='algorithm returns')
        plt.legend(loc=2)
        plt.show()



# -*- encoding:utf-8 -*-
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from os import path
import os
from ..CoreBu.ABuEnv import g_project_data_dir
from ..CoreBu.ABuEnv import g_project_cache_dir
from ..CoreBu.ABuEnv import g_project_rom_data_dir
from ..UtilBu import ABuFileUtil
from ..UtilBu.ABuProgress import AbuProgress

from .ABuXqConsts import columns_map

import pandas as pd

__author__ = '小青蛙'
__weixin__ = 'abu_quant'

# 保存股票信息
STOCK_INFO_DIR = path.join(g_project_data_dir, 'stock_info')
# 保存爬取得股票信息
STOCK_INFO_CACHE_DIR = path.join(g_project_cache_dir, 'stock_info')
STOCK_INFO_ERROR_DIR = path.join(g_project_cache_dir, 'stock_info/error')

ABuFileUtil.ensure_dir(STOCK_INFO_DIR)
ABuFileUtil.ensure_dir(STOCK_INFO_CACHE_DIR)
ABuFileUtil.ensure_dir(STOCK_INFO_ERROR_DIR)


def map_stock_list_rom(market):
    return path.join(g_project_rom_data_dir, 'stock_code_{}.csv'.format(market))


def map_stock_list(market):
    return path.join(STOCK_INFO_DIR, 'stock_code_{}.csv'.format(market))


def map_cache_stock_info(market, symbol):
    """
    临时存取 stock info
    :param market:
    :param symbol:
    :return:
    """
    p = path.join(STOCK_INFO_CACHE_DIR, '{}_{}.txt'.format(market, symbol))
    return p


def save_cache_stock_info(stock_info, market, symbol):
    if stock_info is not None:
        df = pd.DataFrame.from_dict(stock_info, orient='index')
        df.to_csv(map_cache_stock_info(market, symbol))


def exist_stock_info(market, symbol):
    return ABuFileUtil.file_exist(map_cache_stock_info(market, symbol))


def read_stock_symbol_list(market):
    return pd.read_csv(map_stock_list(market))['symbol'].tolist()


def read_stock_symbol_name(market, symbol):
    df = pd.read_csv(map_stock_list(market))
    item = df.loc[df.symbol == symbol]
    return None if item is None else item.name.tolist()[0]


def error_stock_info(market, symbol, error_info):
    """
    抓取stock info 失败 保存到磁盘，以便，下次查看，由于多进程的原因，每个错占用一个空文件,避免多进程共享文件可能出错的问题
    :param error_info:
    :param market:
    :param symbol:
    :return:
    """
    p = path.join(STOCK_INFO_ERROR_DIR, '{}_{}'.format(market, symbol))
    ABuFileUtil.save_file(error_info, p)
    return p


def read_all_error_stock_info_symbol():
    """
    :return: 返回抓取stock info失败的symbol列表
    """
    error_symbols = os.listdir(STOCK_INFO_ERROR_DIR)
    return map(lambda error: error.split('_'), error_symbols)


def query_a_stock(market, symbol):
    stock_df = pd.read_csv(map_stock_list(market))
    return stock_df.loc[stock_df.symbol == symbol]


def _create_a_column(info, columns, size):
    """
    创建size大小的None数组
    :param info:
    :param columns:
    :param size:
    """
    for c in columns:
        if c not in info.keys():
            info[c] = [None] * size


def merge_stock_info_to_stock_list(market=('US', 'HK', 'CN')):
    for m in market:
        stock_df = pd.read_csv(map_stock_list(m), dtype=str)
        extra_info = {}
        # 08012
        progess = AbuProgress(stock_df.shape[0], 0, 'merging {}'.format(m))
        for i, symbol in enumerate(stock_df['symbol']):
            progess.show(i + 1)
            if ABuFileUtil.file_exist(map_cache_stock_info(m, symbol)):
                a_stock_info = pd.read_csv(map_cache_stock_info(m, symbol), dtype=str)

                if a_stock_info is not None and not a_stock_info.empty:
                    keys = a_stock_info.ix[:, 0].tolist()
                    _create_a_column(extra_info, keys, stock_df.shape[0])
                    values = a_stock_info.ix[:, 1].tolist()
                    for k, v in zip(keys, values):
                        extra_info[k][i] = v

        for key in extra_info:
            stock_df[key] = extra_info[key]

        stock_df.fillna('-', inplace=True)
        # 某些symbol的stockinfo为空，，stockinfp为空的原因是stockinfo页面404，因此可以丢弃
        valid_df = stock_df.loc[stock_df.symbol != '-']
        valid_df.to_csv(map_stock_list_rom(m), index=False, encoding='utf-8')


def del_columns(df, columns):
    old_c = df.columns.tolist()
    for col in filter(lambda x: x in old_c, columns):
        df.drop(col, axis=1, inplace=True)


def drop_nuisance(df):
    del_columns(df, ['percent', 'pettm', 'marketcapital', 'volume', 'low52w', 'low', 'high52w', 'high', 'hasexist',
                     'current', 'amount', 'change', 'pe_ttm', 'open', 'last_close', 'chg', 'market_capital',
                     'code'])
    return df


def fix_xq_columns_name():
    """
    雪球获取的数据的key都是中文，dataframe的columns不变与用中文
    """
    for m in ('US', 'CN', 'HK'):
        stock_df = pd.read_csv(map_stock_list_rom(m), dtype=str)
        unnecessary_columns = stock_df.columns.difference(columns_map.keys())
        columns_intersection = stock_df.columns & columns_map.keys()
        del_columns(stock_df, unnecessary_columns)

        stock_df.rename(columns={c: columns_map[c] for c in columns_intersection}, inplace=True)
        stock_df.to_csv(map_stock_list_rom(m), index=True, encoding='utf-8')



# -*- encoding:utf-8 -*-
"""
    买入择时示例因子：黄金分割线买入择时因子
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division


from ..TLineBu import ABuTLGolden
from ..CoreBu.ABuPdHelper import pd_rolling_mean
from .ABuFactorBuyBase import AbuFactorBuyBase, BuyCallMixin

__author__ = '阿布'
__weixin__ = 'abu_quant'


# noinspection PyAttributeOutsideInit
class AbuFactorBuyGolden(AbuFactorBuyBase, BuyCallMixin):
    """示例黄金分割线买入择时因子，混入BuyCallMixin，即向上突破触发买入event"""

    def _init_self(self, **kwargs):
        """
            kwargs中必须: 突破参数xd 比如20，30，40天...作为分割周期
            kwargs中可选参数: 突破参数ma_day(int)，作为在择时突破中的买入event条件
            kwargs中可选参数: 短暂停留阀值stay_day(int)，作为在择时突破中的买入event条件
        """
        # 黄金分割参数xd， 比如20，30，40天...作为分割周期
        self.xd = kwargs['xd']
        # 突破参数ma_day(int), 默认5：代表使用5日移动均线
        self.ma_day = kwargs.pop('ma_day', 5)
        # 短暂停留阀值stay_day(int)，默认进入1天即算
        self.stay_day = kwargs.pop('stay_day', 1)
        # 交易目标短暂停留的标志
        self.below_stay_days = 0
        # 在输出生成的orders_pd中显示的名字
        self.factor_name = '{}:{}'.format(self.__class__.__name__, self.xd)

    def fit_day(self, today):
        """
        策略属于均值回复类型策略：当价格在黄金分割线低档位附近徘徊且有向上回复的趋势时买入目标
        :param today: 当前驱动的交易日金融时间序列数据
        :return:
        """
        # 忽略不符合买入的天（统计周期内前xd天）
        if self.today_ind < self.xd - 1:
            return None

        # 切片从前xd开始直到当前交易日为止的金融序列数据
        window_pd = self.kl_pd[self.today_ind - self.xd + 1: self.today_ind + 1]
        # 在切片的金融序列数据上计算黄金分割档位值，具体阅读ABuTLGolden.calc_golden
        golden = ABuTLGolden.calc_golden(window_pd, show=False)

        """
            目标买入区域在below382－>above382之间，一旦 < below382彻底放弃一段时间的买入观察，
            一旦进入below382－>above382区间也不是立即执行买单，需要在此区域短暂停留后有向上的趋势
            再买入
        """
        if today['close'] < golden.below382:
            # 进入< below382 彻底放弃一段时间的买入观察，放弃周期＝self.xd
            self.below_stay_days = 0
            self.skip_days = self.xd
        elif golden.above382 > today['close'] >= golden.below382:
            # 作为交易目标短暂停留的标志
            self.below_stay_days += 1

        def ma_break_func():
            """
               分别计算xd天的移动平均价格和初始化中的ma_day天的移动平均价格，
               买入条件是ma_xd > ma_cross, 即快线突破慢线
            """
            # 快线尾巴
            ma_cross = pd_rolling_mean(window_pd.close, window=self.ma_day, min_periods=self.ma_day)[-1]
            # 慢线尾巴
            ma_xd = pd_rolling_mean(window_pd.close, window=self.xd, min_periods=self.xd)[-1]
            if ma_cross > ma_xd:
                # 快线突破慢线
                return True
            return False

        """
            self.below_stay_days >= self.stay_day: 代表进入below382－>above382区域短暂停留，
                                                   可以通过_init_self参数微调短暂停留阀值stay_day
            today['close'] >= golden.above382 or ma_break_func(): 代表有向上的趋势
        """
        if self.below_stay_days >= self.stay_day and (today['close'] >= golden.above382 or ma_break_func()):
            # 重置交易目标短暂停留的标志
            self.below_stay_days = 0
            # 放弃一段时间的买入观察, 放弃周期＝self.xd/2
            self.skip_days = self.xd / 2
            # 生成买单, 由于使用了今天的收盘价格做为策略信号判断，所以信号发出后，只能明天买
            return self.buy_tomorrow()
        return None



# -*- encoding:utf-8 -*-
"""
    选股具体工作者，整合金融时间序列，选股因子，资金类进行
    选股操作，在择时金融时间序列之前一段时间上迭代初始交易对象
    进行选股因子的拟合操作
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import copy

from .ABuPickBase import AbuPickStockWorkBase
from ..MarketBu.ABuMarket import all_symbol
from ..PickStockBu.ABuPickStockBase import AbuPickStockBase
from ..UtilBu.ABuProgress import AbuMulPidProgress

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuPickStockWorker(AbuPickStockWorkBase):
    """选股类"""

    def __init__(self, capital, benchmark, kl_pd_manager, choice_symbols=None, stock_pickers=None):
        """
        :param capital: 资金类AbuCapital实例化对象
        :param benchmark: 交易基准对象，AbuBenchmark实例对象
        :param kl_pd_manager: 金融时间序列管理对象，AbuKLManager实例
        :param choice_symbols: 初始备选交易对象序列
        :param stock_pickers: 选股因子序列
        """
        self.capital = capital
        self.benchmark = benchmark
        self.choice_symbols = choice_symbols
        self.kl_pd_manager = kl_pd_manager
        self.stock_pickers = []
        self.first_stock_pickers = []
        self.init_stock_pickers(stock_pickers)

    def __str__(self):
        """打印对象显示：选股因子序列＋选股交易对象"""
        return 'stock_pickers:{}\nchoice_symbols:{}'.format(self.stock_pickers, self.choice_symbols)

    __repr__ = __str__

    def init_stock_pickers(self, stock_pickers):
        """
        通过stock_pickers实例化各个选股因子
        :param stock_pickers:list中元素为dict，每个dict为因子的构造元素，如class，构造参数等
        :return:
        """
        if stock_pickers is not None:
            for picker_class in stock_pickers:
                if picker_class is None:
                    continue

                if 'class' not in picker_class:
                    # 必须要有需要实例化的类信息
                    raise ValueError('picker_class class key must name class !!!')

                picker_class_cp = copy.deepcopy(picker_class)
                # pop出类信息后剩下的都为类需要的参数
                class_fac = picker_class_cp.pop('class')
                # 整合capital，benchmark等实例化因子对象
                picker = class_fac(self.capital, self.benchmark, **picker_class_cp)

                if not isinstance(picker, AbuPickStockBase):
                    # 因子对象类型检测
                    raise TypeError('factor must base AbuPickStockBase')

                if 'first_choice' in picker_class and picker_class['first_choice']:
                    # 如果参数设置first_choice且是True, 添加到first_stock_pickers选股序列
                    self.first_stock_pickers.append(picker)
                else:
                    self.stock_pickers.append(picker)
        if self.choice_symbols is None or len(self.choice_symbols) == 0:
            # 如果参数中初始备选交易对象序列为none, 从对应市场中获取所有的交易对象，详情查阅all_symbol
            self.choice_symbols = all_symbol()

    def fit(self):
        """
        选股开始工作，与择时不同，选股是数据多对多，
        即多个交易对象对多个选股因子配合资金基准等参数工作
        :return:
        """

        def _first_batch_fit():
            """
            first_choice选股：针对备选池进行选股，迭代选股因子，使用因子的fit_first_choice方法
            即因子内部提供批量选股高效的首选方法
            :return:
            """
            if self.first_stock_pickers is None or len(self.first_stock_pickers) == 0:
                # 如果没有first_stock_picker要返回self.choice_symbols，代表没有投任何反对票，全部通过
                return self.choice_symbols

            # 首选将所有备选对象赋予inner_first_choice_symbols
            inner_first_choice_symbols = self.choice_symbols
            with AbuMulPidProgress(len(self.first_stock_pickers), 'pick first_choice stocks complete') as progress:
                for epoch, first_choice in enumerate(self.first_stock_pickers):
                    progress.show(epoch + 1)
                    # 每一个选股因子通过fit_first_choice对inner_first_choice_symbols进行筛选，滤网一层一层过滤
                    inner_first_choice_symbols = first_choice.fit_first_choice(self, inner_first_choice_symbols)
            return inner_first_choice_symbols

        def _batch_fit():
            """
            普通选股：针对备选池进行选股，迭代初始选股序列，在迭代中再迭代选股因子，选股因子决定是否对
            symbol投出反对票，一旦一个因子投出反对票，即筛出序列，一票否决
            :return:
            """
            if self.stock_pickers is None or len(self.stock_pickers) == 0:
                # 如果没有stock_pickers要返回self.choice_symbols，代表没有投任何反对票，全部通过
                return self.choice_symbols

            with AbuMulPidProgress(len(self.choice_symbols), 'pick stocks complete') as progress:
                # 启动选股进度显示
                inner_choice_symbols = []
                for epoch, target_symbol in enumerate(self.choice_symbols):
                    progress.show(epoch + 1)

                    add = True
                    for picker in self.stock_pickers:
                        kl_pd = self.kl_pd_manager.get_pick_stock_kl_pd(target_symbol, picker.xd, picker.min_xd)
                        if kl_pd is None:
                            # 注意get_pick_stock_kl_pd内部对选股金融序列太少的情况进行过滤，详情get_pick_stock_kl_pd
                            add = False
                            break
                        sub_add = picker.fit_pick(kl_pd, target_symbol)
                        if sub_add is False:
                            # 只要一个选股因子投了反对票，就刷出
                            add = False
                            break
                    if add:
                        inner_choice_symbols.append(target_symbol)
                return inner_choice_symbols

        # 筛选各个因first_choice序列，返回给self.choice_symbols，_batch_fit继续晒
        self.choice_symbols = _first_batch_fit()
        # 通过两次迭代继续筛选
        self.choice_symbols = _batch_fit()



# -*- encoding:utf-8 -*-
"""卖出因子参数以及选择图形可视化"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import ipywidgets as widgets

from ..FactorSellBu.ABuFactorAtrNStop import AbuFactorAtrNStop
from ..FactorSellBu.ABuFactorCloseAtrNStop import AbuFactorCloseAtrNStop
from ..FactorSellBu.ABuFactorPreAtrNStop import AbuFactorPreAtrNStop
from ..FactorSellBu.ABuFactorSellBreak import AbuFactorSellBreak
from ..FactorSellBu.ABuFactorSellDM import AbuDoubleMaSell
from ..FactorSellBu.ABuFactorSellNDay import AbuFactorSellNDay
from ..WidgetBu.ABuWGSFBase import WidgetFactorSellBase

__author__ = '阿布'
__weixin__ = 'abu_quant'


class SellDMWidget(WidgetFactorSellBase):
    """对应AbuDoubleMaSell策略widget"""

    def _init_widget(self):
        """构建AbuDoubleMaSell策略参数界面"""

        self.description = widgets.Textarea(
            value=u'双均线卖出策略：\n'
                  u'双均线策略是量化策略中经典的策略之一，其属于趋势跟踪策略: \n'
                  u'1. 预设两条均线：如一个ma=5，一个ma=60, 5的均线被称作快线，60的均线被称作慢线\n'
                  u'2. 择时卖出策略中当快线下穿慢线（ma5下穿ma60）称为形成死叉卖点信号，卖出股票\n',
            description=u'双均线卖',
            disabled=False,
            layout=self.description_layout
        )

        self.slow_label = widgets.Label(u'默认慢线ma=60:当快线下穿慢线称为形成死叉', layout=self.label_layout)
        self.slow_int = widgets.IntSlider(
            value=60,
            min=10,
            max=120,
            step=1,
            description=u'慢线',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='d'
        )
        self.slow_box = widgets.VBox([self.slow_label, self.slow_int])

        self.fast_label = widgets.Label(u'默认快线ma=5:当快线下穿慢线称为形成死叉', layout=self.label_layout)
        self.fast_int = widgets.IntSlider(
            value=5,
            min=1,
            max=90,
            step=1,
            description=u'快线',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='d'
        )
        self.fast_box = widgets.VBox([self.fast_label, self.fast_int])
        self.widget = widgets.VBox([self.description, self.slow_box, self.fast_box, self.add_box],
                                   # border='solid 1px',
                                   layout=self.widget_layout)

    def make_sell_factor_unique(self):
        """对应按钮添加AbuDoubleMaSell策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        factor_dict = {'class': AbuDoubleMaSell, 'slow': self.slow_int.value, 'fast': self.fast_int.value}
        factor_desc_key = u'动态双均慢{}快{}卖出'.format(self.slow_int.value, self.fast_int.value)
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuDoubleMaSell"""
        return AbuDoubleMaSell


class SellAtrNStopWidget(WidgetFactorSellBase):
    """对应AbuFactorAtrNStop策略widget"""

    def _init_widget(self):
        """构建AbuFactorAtrNStop策略参数界面"""

        self.description = widgets.Textarea(
            value=u'止盈策略 & 止损策略：\n'
                  u'1. 真实波幅atr作为最大止盈和最大止损的常数值\n'
                  u'2. 当stop_loss_n 乘以 当日atr > 买入价格 － 当日收盘价格->止损卖出\n'
                  u'3. 当stop_win_n 乘以 当日atr < 当日收盘价格 －买入价格->止盈卖出',
            description=u'止盈止损',
            disabled=False,
            layout=self.description_layout
        )
        self.stop_loss_n_label = widgets.Label(u'stop_loss_n乘以当日atr大于买价减close->止损',
                                               layout=self.label_layout)
        self.stop_loss_n = widgets.FloatSlider(
            value=1.0,
            min=0.10,
            max=10.0,
            step=0.1,
            description='stop_loss_n',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='.1f',
        )
        self.stop_loss_n_box = widgets.VBox([self.stop_loss_n_label, self.stop_loss_n])
        self.stop_win_n_label = widgets.Label(u'stop_win_n乘以当日atr小于close减买价->止盈',
                                              layout=self.label_layout)
        self.stop_win_n = widgets.FloatSlider(
            value=3.0,
            min=0.10,
            max=10.0,
            step=0.10,
            description='stop_win_n',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='.1f',
        )
        self.stop_win_n_box = widgets.VBox([self.stop_win_n_label, self.stop_win_n])
        self.widget = widgets.VBox([self.description, self.stop_loss_n_box,
                                    self.stop_win_n_box, self.add_box],  # border='solid 1px',
                                   layout=self.widget_layout)

    def make_sell_factor_unique(self):
        """对应按钮添加AbuFactorAtrNStop策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        factor_dict = {'class': AbuFactorAtrNStop,
                       'stop_win_n': self.stop_win_n.value,
                       'stop_loss_n': self.stop_loss_n.value}
        factor_desc_key = u'n atr止盈{}止损{}'.format(self.stop_win_n.value, self.stop_loss_n.value)
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuFactorAtrNStop"""
        return AbuFactorAtrNStop


class SellCloseAtrNWidget(WidgetFactorSellBase):
    """对应AbuFactorCloseAtrNStop策略widget"""

    def _init_widget(self):
        """构建AbuFactorCloseAtrNStop策略参数界面"""
        self.description = widgets.Textarea(
            value=u'利润保护止盈策略：\n'
                  u'1. 买入后最大收益价格 - 今日价格 > 一定收益\n'
                  u'2. 买入后最大收益价格 - 今日价格 < close_atr_n * 当日atr\n'
                  u'3. 当买入有一定收益后，如果下跌幅度超过close_atr_n乘以当日atr->保护止盈卖出',
            description=u'保护止盈',
            disabled=False,
            layout=self.description_layout
        )

        self.close_atr_n_label = widgets.Label(u'收益下跌超过close_atr_n乘以当日atr->保护止盈',
                                               layout=self.label_layout)
        self.close_atr_n = widgets.FloatSlider(
            value=1.5,
            min=0.10,
            max=10.0,
            step=0.1,
            description='close_atr_n',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='.1f',
        )
        self.close_atr_n_box = widgets.VBox([self.close_atr_n_label, self.close_atr_n])
        self.widget = widgets.VBox([self.description, self.close_atr_n_box, self.add_box],  # border='solid 1px',
                                   layout=self.widget_layout)

    def make_sell_factor_unique(self):
        """对应按钮添加AbuFactorCloseAtrNStop策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        factor_dict = {'class': AbuFactorCloseAtrNStop,
                       'close_atr_n': self.close_atr_n.value}
        factor_desc_key = u'利润保护止盈n={}'.format(self.close_atr_n.value)
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuFactorCloseAtrNStop"""
        return AbuFactorCloseAtrNStop


class SellPreAtrNWidget(WidgetFactorSellBase):
    """对应AbuFactorPreAtrNStop策略widget"""

    def _init_widget(self):
        """构建AbuFactorPreAtrNStop策略参数界面"""
        self.description = widgets.Textarea(
            value=u'风险控制止损策略：\n'
                  u'1. 单日最大跌幅n倍atr止损\n'
                  u'2. 当今日价格下跌幅度 > 当日atr 乘以 pre_atr_n（下跌止损倍数）卖出操作',
            description=u'风险止损',
            disabled=False,
            layout=self.description_layout
        )

        self.pre_atr_n_label = widgets.Label(u'当今天价格开始剧烈下跌，采取果断平仓措施',
                                             layout=self.label_layout)
        self.pre_atr_n = widgets.FloatSlider(
            value=1.5,
            min=0.10,
            max=10.0,
            step=0.1,
            description='pre_atr_n',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='.1f',
        )
        self.pre_atr_n_box = widgets.VBox([self.pre_atr_n_label, self.pre_atr_n])
        self.widget = widgets.VBox([self.description, self.pre_atr_n_box, self.add_box],  # border='solid 1px',
                                   layout=self.widget_layout)

    def make_sell_factor_unique(self):
        """对应按钮添加AbuFactorPreAtrNStop策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        factor_dict = {'class': AbuFactorPreAtrNStop,
                       'pre_atr_n': self.pre_atr_n.value}
        factor_desc_key = u'风险控制止损n={}'.format(self.pre_atr_n.value)
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuFactorPreAtrNStop"""
        return AbuFactorPreAtrNStop


class SellXDWidget(WidgetFactorSellBase):
    """对应AbuFactorSellBreak策略widget"""

    def _init_widget(self):
        """构建AbuFactorSellBreak策略参数界面"""
        self.description = widgets.Textarea(
            value=u'海龟向下趋势突破卖出策略：\n'
                  u'趋势突破定义为当天收盘价格低于N天内的最低价，作为卖出信号，卖出操作',
            description=u'海龟卖出',
            disabled=False,
            layout=self.description_layout
        )
        self.xd_label = widgets.Label(u'突破周期参数：比如21，30，42天....突破',
                                      layout=self.label_layout)
        self.xd = widgets.IntSlider(
            value=10,
            min=3,
            max=120,
            step=1,
            description=u'周期',
            disabled=False,
            orientation='horizontal',
            readout=True,
            readout_format='d'
        )
        self.xd_box = widgets.VBox([self.xd_label, self.xd])
        self.widget = widgets.VBox([self.description, self.xd_box, self.add_box],  # border='solid 1px',
                                   layout=self.widget_layout)

    def make_sell_factor_unique(self):
        """对应按钮添加AbuFactorSellBreak策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        factor_dict = {'class': AbuFactorSellBreak, 'xd': self.xd.value}
        factor_desc_key = u'海龟{}天趋势突破卖出'.format(self.xd.value)
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuFactorSellBreak"""
        return AbuFactorSellBreak


class SellNDWidget(WidgetFactorSellBase):
    """对应AbuFactorSellNDay策略widget"""

    def _init_widget(self):
        """构建AbuFactorSellNDay策略参数界面"""
        self.description = widgets.Textarea(
            value=u'持有N天后卖出策略：\n'
                  u'卖出策略，不管交易现在什么结果，买入后只持有N天\n'
                  u'需要与特定\'买入策略\'形成配合\n，'
                  u'单独使用N天卖出策略意义不大',
            description=u'N天卖出',
            disabled=False,
            layout=self.description_layout
        )
        sell_n_label = widgets.Label(u'设定买入后只持有天数，默认1', layout=self.label_layout)
        self.sell_n = widgets.IntText(
            value=1,
            description=u'N天',
            disabled=False
        )
        sell_n_box = widgets.VBox([sell_n_label, self.sell_n])

        is_sell_today_label = widgets.Label(u'设定买入n天后，当天还是隔天卖出', layout=self.label_layout)
        self.is_sell_today = widgets.Dropdown(
            options={u'N天后隔天卖出': False, u'N天后当天卖出': True},
            value=False,
            description=u'当天隔天:',
        )
        is_sell_today_box = widgets.VBox([is_sell_today_label, self.is_sell_today])

        self.widget = widgets.VBox([self.description, sell_n_box, is_sell_today_box, self.add_box],
                                   # border='solid 1px',
                                   layout=self.widget_layout)

    def make_sell_factor_unique(self):
        """对应按钮添加AbuFactorSellNDay策略，构建策略字典对象factor_dict以及唯一策略描述字符串factor_desc_key"""
        factor_dict = {'class': AbuFactorSellNDay, 'sell_n': self.sell_n.value,
                       'is_sell_today': self.is_sell_today.value}
        factor_desc_key = u'持有{}天{}卖出'.format(self.sell_n.value, u'当天' if self.is_sell_today.value else u'隔天')
        return factor_dict, factor_desc_key

    def delegate_class(self):
        """子类因子所委托的具体因子类AbuFactorSellNDay"""
        return AbuFactorSellNDay



# -*- encoding:utf-8 -*-
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from abc import ABCMeta, abstractmethod

import logging
import time

from . import ABuXqFile

from ..CoreBu import env
# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import map, reduce, filter
from .ABuXqApi import BASE_XQ_HQ_URL
from .ABuXqApi import BASE_XQ_STOCK_INFO
from ..ExtBu import six

__author__ = '小青蛙'
__weixin__ = 'abu_quant'


def _bs4_html(content):
    """
    使用BeautifulSoup解析html
    :param content: html
    :return: BeautifulSoup
    """
    from bs4 import BeautifulSoup
    return BeautifulSoup(content, 'lxml')


def _xpath(content):
    """
    使用xpath解析html 
    :param content:
    :return:
    """
    from lxml import etree
    selector = etree.HTML(content)
    return selector


class BaseXQCrawlBrower(six.with_metaclass(ABCMeta, object)):
    """
    使用chrome浏览器的自动化测试驱动接口，获取网页数据
    """

    def __init__(self, base_url):
        self._base_url = base_url
        if env.g_crawl_chrome_driver is not None:
            self.driver_path = env.g_crawl_chrome_driver
        else:
            raise RuntimeError('driver_path error!!!, abupy.CoreBu.ABuEnv.g_crawl_chrome_driver must be right')

        # noinspection PyUnresolvedReferences
        from selenium.webdriver.support import ui
        # noinspection PyUnresolvedReferences
        from selenium import webdriver
        self.driver = webdriver.Chrome(self.driver_path)
        self.wait = ui.WebDriverWait(self.driver, 10)

    @abstractmethod
    def _crawl_imp(self, *args, **kwargs):
        pass

    def get(self, url):
        self.driver.get(url)

    @property
    def content(self):
        return self.driver.page_source

    def crawl(self, *args, **kwargs):
        """
        执行完任务是自动退出，避免占用资源，在多进程爬时会启动多个chrome实例
        :param args:
        :param kwargs:
        :return: crawl_imp
        """
        ret = None
        try:
            self.driver.get(self._base_url)
            self.driver.maximize_window()
            ret = self._crawl_imp(*args, **kwargs)
        except Exception as e:
            logging.exception(e)
        return ret

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.driver.quit()

    def _scroll_to_bottom(self, element):
        loc = element.location
        self.driver.execute_script('window.scrollBy({},{})'.format(loc['x'], loc['y']))


class BaseHQCrawlBrower(BaseXQCrawlBrower):
    def _crawl_imp(self, *args, **kwargs):
        pass

    def __init__(self, url):
        super(BaseHQCrawlBrower, self).__init__(BASE_XQ_HQ_URL)
        self._base_url = self._base_url + url


class NavHQCrawlBrower(BaseHQCrawlBrower):
    def _crawl_imp(self, *args, **kwargs):
        return _parse_nav(self.driver.page_source)

    def __init__(self):
        super(NavHQCrawlBrower, self).__init__('')


class StockListCrawlBrower(BaseHQCrawlBrower):
    def _ensure_max_page_size(self):
        """
        使每页展示的stock数最多，总页数变少，使网络请求数变少
        """
        max_page_tag = self.driver.find_element_by_xpath('//*[@id="stockList-header"]/div[2]/a[3]')
        max_page_tag.click()
        time_out = 30
        while time_out:
            time.sleep(1)
            time_out -= 1
            _, total = self._curr_total_page()
            #  直到 最大size生效
            if total == 1 or self._curr_page_counts() == int(max_page_tag.text):
                break

    def _curr_page_counts(self):
        selector = _xpath(self.content)
        items = selector.xpath('//*[@id="stockList"]/div[1]/table/tbody/tr')
        return len(items)

    def _curr_total_page(self):

        selector = _xpath(self.content)
        pages = selector.xpath('//*[@id="pageList"]/div/ul/li/a/text()')
        cur_page = selector.xpath('//*[@id="pageList"]/div/ul/li[@class="active"]/a/text()')

        # 存在pages的最后一个值，否则cur和total都是1
        if len(pages):
            return int(cur_page[0]), int(pages[-1])
        else:
            return 1, 1

    def _curr_page_items(self):
        selector = _xpath(self.content)
        # code = selector.xpath('//*[@id="stockList"]/div[1]/table/tbody/tr/td[1]/a/text()')
        # name = selector.xpath('//*[@id="stockList"]/div[1]/table/tbody/tr/td[2]/a/text()')
        # a标签下的text() 可能不存在，而xpath会把不存在的过滤掉，导致code，和name的长度不一致，产生错位，故先找到a，a。text为kong也占位，就能一一对应
        code = selector.xpath('//*[@id="stockList"]/div[1]/table/tbody/tr/td[1]/a')
        name = selector.xpath('//*[@id="stockList"]/div[1]/table/tbody/tr/td[2]/a')
        code = list(map(lambda a: a.text, code))
        name = list(map(lambda a: a.text, name))
        return name, code

    def _goto_next_page(self):
        next_page = self.driver.find_element_by_xpath('//*[@id="pageList"]/div/ul/li[@class="next"]/a')
        if next_page is not None:
            # 滚动到next_page 标签显示出来，否则click可能会报错
            self.wait.until(lambda dr: next_page.is_enabled())
            self.driver.execute_script('arguments[0].click()', next_page)
            time.sleep(1)

    def _crawl_imp(self, *args, **kwargs):
        self._ensure_max_page_size()

        cur_page, total_page = self._curr_total_page()
        names = []
        symbols = []
        # page index start 1
        for page in range(1, total_page + 1):
            self.wait.until(lambda dr: dr.find_element_by_xpath('//*[@id="stockList"]/div[1]/table').is_displayed())
            cur_page, _ = self._curr_total_page()
            temp_names, temp_symbols = self._curr_page_items()
            names += temp_names
            symbols += temp_symbols
            if page < total_page:
                self._goto_next_page()
            else:
                break

        return names, symbols

    def __init__(self, url):
        super(StockListCrawlBrower, self).__init__(url)


def _parse_nav(content):
    soup = _bs4_html(content)
    nav_tags = soup.select('.industry-nav > div')

    def parse_nav_tag(tag):
        nav = {}
        first_nav = tag.select('.first-nav > span')
        if len(first_nav) > 0:
            second_nav = tag.select('.second-nav > li')
            nav[first_nav[0].string] = {}
            for nav_2 in second_nav:
                a_tag = nav_2.select('a')
                if len(a_tag) <= 0:
                    continue
                third_nav = nav_2.select('.third-nav')
                if len(third_nav) > 0:
                    second_nav_name = str(nav_2).replace('<li><i class="list-style"></i>', '')
                    second_nav_name = second_nav_name[: second_nav_name.index('<i')]
                    nav[first_nav[0].string][second_nav_name] = list(map(lambda a: {a.get('title'): a.get('href')},
                                                                         a_tag))
                else:
                    nav[first_nav[0].string][a_tag[0].get('title')] = a_tag[0].get('href')
        return nav

    def merge(dict1, dict2):
        return dict(dict1, **dict2)

    return reduce(lambda d1, d2: merge(d1, d2), map(lambda tag: parse_nav_tag(tag), nav_tags))


class StockInfoListBrower(BaseXQCrawlBrower):
    def __init__(self, market, symbols):
        super(StockInfoListBrower, self).__init__(BASE_XQ_STOCK_INFO)

        self._market = market
        self._symbols = symbols

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._market = None
        self._symbols = None
        super(StockInfoListBrower, self).__exit__(exc_type, exc_val, exc_tb)

    def _parse_stock_info(self):
        selector = _xpath(self.content)
        # 特斯拉(NASDAQ:TSLA)
        stock_name = selector.xpath('//*[@id="center"]/div[2]/div[2]/div[1]/div[1]/span[1]/strong/text()')
        company_info_p = selector.xpath('//*[@id="center"]/div[3]/div/div[2]/div/p')
        company_industry = selector.xpath('//*[@id="relatedIndustry"]/h2/a/text()')

        quate_items = selector.xpath('//*[@id="center"]/div[2]/div[2]/div[2]/table/tbody/tr/td')

        info = {}
        if len(stock_name):
            stock_name_info = stock_name[0]
            st = stock_name_info.rfind('(')
            en = stock_name_info.rfind(')')
            market_symbol = stock_name_info[st + 1:en]
            sp_result = market_symbol.split(':')
            if len(sp_result) == 2:
                company_name = stock_name_info[:st]
                exchange_name = sp_result[0]
                company_symbol = sp_result[1]
                info['name'] = company_name
                info['exchange'] = exchange_name
                info['symbol'] = company_symbol

        if len(company_info_p):
            last_key = None
            for p in company_info_p:
                for child in p.getchildren():
                    if child.tag == 'strong':
                        info[child.text] = child.tail
                        last_key = child.text
                    if child.tag == 'a':
                        info[last_key] = child.get('href')
        if len(quate_items):
            for item in quate_items:
                if len(item.getchildren()) == 1:
                    info[item.text] = item.getchildren()[0].text

        if len(company_industry):
            info['industry'] = company_industry[0]

        return info

    def _crawl_imp(self, *args, **kwargs):
        for index, symbol in enumerate(self._symbols):
            try:
                if not ABuXqFile.exist_stock_info(self._market, symbol) or ('replace' in kwargs and kwargs['replace']):
                    self.get(self._base_url + symbol)
                    stock_info = self._parse_stock_info()
                    ABuXqFile.save_cache_stock_info(stock_info, self._market, symbol)
                if (index + 1) % 200 == 0:
                    print(
                        '{}: {} {}  {}/{}'.format(kwargs['process'], self._market, symbol, index + 1,
                                                  len(self._symbols)))
            except Exception as e:
                # 记录失败的symbol
                ABuXqFile.error_stock_info(self._market, symbol, e)
                logging.exception(e)
        return 'Done'



# coding=utf-8
"""
    市场相关切割，选股，等操作模块
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import datetime

import numpy as np
import pandas as pd

from ..UtilBu import ABuFileUtil
from ..CoreBu import ABuEnv
from ..CoreBu.ABuDeprecated import AbuDeprecated
from ..CoreBu.ABuEnv import EMarketTargetType, EMarketSubType
# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import filter, zip
from ..CoreBu.ABuFixes import KFold, six
from ..UtilBu.ABuLazyUtil import LazyFunc
from ..MarketBu.ABuSymbol import Symbol, code_to_symbol
from ..MarketBu.ABuSymbolFutures import AbuFuturesCn, AbuFuturesGB
from ..MarketBu.ABuSymbolStock import AbuSymbolCN, AbuSymbolUS, AbuSymbolHK

__author__ = '阿布'
__weixin__ = 'abu_quant'

# TODO 在全市场回测时设置g_use_env_market_set=True
"""默认False，如果进行全市场回测这里可以设置True, 配合LazyFunc提高效率"""
g_use_env_market_set = False

"""在market_train_test_split函数中，切割的测试集交易symbol，本地序列化存储路径的基础路径名"""
K_MARKET_TEST_FN_BASE = os.path.join(ABuEnv.g_project_cache_dir, 'market_test_symbols')
"""在market_train_test_split函数中，切割的训练集交易symbol，本地序列化存储路径的基础路径名"""
K_MARKET_TRAIN_FN_BASE = os.path.join(ABuEnv.g_project_cache_dir, 'market_train_symbols')

# TODO 从沙盒数据库里读取，否则之后有变动还需要跟着改
K_SAND_BOX_US = ['usTSLA', 'usNOAH', 'usSFUN', 'usBIDU', 'usAAPL', 'usGOOG', 'usWUBA', 'usVIPS']
K_SAND_BOX_CN = ['sz002230', 'sz300104', 'sz300059', 'sh601766', 'sh600085', 'sh600036',
                 'sh600809', 'sz000002', 'sz002594', 'sz002739']
K_SAND_BOX_HK = ['hk03333', 'hk00700', 'hk02333', 'hk01359', 'hk00656', 'hk03888', 'hk02318']


# noinspection PyUnresolvedReferences
class MarketMixin(object):
    """
        市场信息混入类，被混入类需要设置self.symbol_name，
        通过code_to_symbol将symbol转换为Symbol对象, 通过Symbol对象
        查询market和sub_market
    """

    @LazyFunc
    def _symbol(self):
        """通过code_to_symbol将symbol转换为Symbol对象 LazyFunc"""
        if not hasattr(self, 'symbol_name'):
            # 被混入类需要设置self.symbol_name
            raise NameError('must name symbol_name!!')
        # 通过code_to_symbol将symbol转换为Symbol对象
        return code_to_symbol(self.symbol_name)

    @LazyFunc
    def symbol_market(self):
        """查询self.symbol_name的市场类型 LazyFunc"""
        return self._symbol.market

    @LazyFunc
    def symbol_sub_market(self):
        """查询self.symbol_name的子市场类型，即交易所信息 LazyFunc"""
        return self._symbol.sub_market


def split_k_market(k_split, market_symbols=None, market=None):
    """
    将market_symbols序列切分成k_split个序列
    :param k_split: 切分成子序列个数int
    :param market_symbols: 待切割的原始symbols序列，如果none, 将取market参数中指定的市场所有symbol
    :param market: 默认None，如None则服从ABuEnv.g_market_target市场设置
    :return: list序列，序列中的每一个元素都是切割好的子symbol序列
    """
    if market is None:
        # 取env中的ABuEnv.g_market_target设置
        market = ABuEnv.g_market_target
    if market_symbols is None:
        # 取market参数中指定的市场所有symbol
        market_symbols = all_symbol(market=market)
    if len(market_symbols) < k_split:
        # 特殊情况，eg：k_split＝100，但是len(market_symbols)＝50，就不切割了，直接返回
        return [[symbol] for symbol in market_symbols]

    # 计算每一个子序列的承载的symbol个数，即eg：100 ／ 5 ＝ 20
    sub_symbols_cnt = int(len(market_symbols) / k_split)
    group_adjacent = lambda a, k: zip(*([iter(a)] * k))
    # 使用lambda函数group_adjacent将market_symbols切割子序列，每个子系列sub_symbols_cnt个
    symbols = list(group_adjacent(market_symbols, sub_symbols_cnt))
    # 将不能整除的余数symbol个再放进去
    residue_ind = -(len(market_symbols) % sub_symbols_cnt) if sub_symbols_cnt > 0 else 0
    if residue_ind < 0:
        # 所以如果不能除尽，最终切割的子序列数量为k_split+1, 外部如果需要进行多认为并行，可根据最终切割好的数量重分配任务数
        symbols.append(market_symbols[residue_ind:])
    return symbols


def choice_symbols(count, market_symbols=None, market=None):
    """
    在market_symbols中随机选择count个symbol，不放回随机的抽取方式
    :param count: 选择count个(int)
    :param market_symbols: 备选symbols序列，如果None, 则从参数market选择全市场symbol做为备选
    :param market: 默认None，如None则服从ABuEnv.g_market_target市场设置
    :return: 随机选择count个symbol结果序列
    """
    if market is None:
        # 如None则服从ABuEnv.g_market_target市场设置
        market = ABuEnv.g_market_target
    if market_symbols is None:
        # 从参数market选择全市场symbol做为备选
        market_symbols = all_symbol(market=market)
    # 使用不放回随机的抽取方式
    return np.random.choice(market_symbols, count, replace=False)


def choice_symbols_with_replace(count, market_symbols=None, market=None):
    """
    在market_symbols中随机选择count个symbol，有放回随机的抽取方式
    :param count: 选择count个(int)
    :param market_symbols: 备选symbols序列，如果None, 则从参数market选择全市场symbol做为备选
    :param market: 默认None，如None则服从ABuEnv.g_market_target市场设置
    :return: 随机选择count个symbol结果序列
    """

    if market is None:
        market = ABuEnv.g_market_target
    if market_symbols is None:
        market_symbols = all_symbol(market=market)
    # 使用有放回随机的抽取方式，即replace=True
    return np.random.choice(market_symbols, count, replace=True)


def _all_us_symbol(index=False):
    """
    通过AbuSymbolUS().all_symbol获取美股全市场股票代码
    :param index: 是否包含指数
    :return:
    """

    # noinspection PyProtectedMember
    if ABuEnv._g_enable_example_env_ipython:
        return K_SAND_BOX_US
    return AbuSymbolUS().all_symbol(index=index)


def _all_cn_symbol(index=False):
    """
    通过AbuSymbolCN().all_symbol获取A股全市场股票代码
    :param index: 是否包含指数
    :return:
    """
    # noinspection PyProtectedMember
    if ABuEnv._g_enable_example_env_ipython:
        return K_SAND_BOX_CN
    return AbuSymbolCN().all_symbol(index=index)


def _all_hk_symbol(index=False):
    """
    通过AbuSymbolHK().all_symbol获取A股全市场股票代码
    :param index: 是否包含指数
    :return:
    """
    # noinspection PyProtectedMember
    if ABuEnv._g_enable_example_env_ipython:
        return K_SAND_BOX_HK
    return AbuSymbolHK().all_symbol(index=index)


def _all_futures_cn():
    """
    AbuFuturesCn().symbol获取国内期货symbol代码，注意这里只取连续合约代码
    :return:
    """
    return AbuFuturesCn().symbol


def _all_futures_gb():
    """
    AbuFuturesGB().symbol获取国际期货symbol代码，LME，CBOT，COMEX
    :return:
    """
    return AbuFuturesGB().symbol


def _all_tc_symbol():
    """
    获取币类symbol，注意这里只取比特币与莱特币，可自行扩展其它币种
    :return:
    """
    return ['btc', 'ltc']


def all_symbol(market=None, ss=False, index=False, value=True):
    """
    根据传入的市场获取全市场代码
    :param market: 默认None，如None则服从ABuEnv.g_market_target市场设置
    :param ss: 是否将返回序列使用pd.Series包装
    :param index: 是否包含指数大盘symbol
    :param value: 返回字符串值，即如果序列中的元素是Symbol对象，Symbol转换字符串
    :return:
    """
    if market is None:
        # None则服从ABuEnv.g_market_target市场设置
        market = ABuEnv.g_market_target

    if market == EMarketTargetType.E_MARKET_TARGET_US:
        symbols = _all_us_symbol(index)
    elif market == EMarketTargetType.E_MARKET_TARGET_CN:
        symbols = _all_cn_symbol(index)
    elif market == EMarketTargetType.E_MARKET_TARGET_HK:
        symbols = _all_hk_symbol(index)
    elif market == EMarketTargetType.E_MARKET_TARGET_FUTURES_CN:
        symbols = _all_futures_cn()
    elif market == EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL:
        symbols = _all_futures_gb()
    elif market == EMarketTargetType.E_MARKET_TARGET_TC:
        symbols = _all_tc_symbol()
    else:
        raise TypeError('JUST SUPPORT EMarketTargetType!')

    # 在出口统一确保唯一性, 在每一个内部_all_xx_symbol中也要尽量保证唯一
    symbols = list(set(symbols))

    if value:
        """
            如果是Symbol类型的还原成字符串，尽量在上面返回的symbols序列是字符串类型
            不要在上面构造symbol，浪费效率，统一会在之后的逻辑中通过code_to_symbol
            进行转换
        """
        symbols = [sb.value if isinstance(sb, Symbol) else sb for sb in symbols]
    # 根据参数ss是否将返回序列使用pd.Series包装
    return pd.Series(symbols) if ss else symbols


def query_symbol_market(target_symbol):
    """
    查询target_symbol所对应的市场对象EMarketTargetType
    :param target_symbol: 支持Symbol对象类型和字符串对象类型
    :return: EMarketTargetType对象
    """
    symbol_obj = None
    if target_symbol is None:
        symbol_obj = None
    elif isinstance(target_symbol, Symbol):
        symbol_obj = target_symbol
    elif isinstance(target_symbol, six.string_types):
        try:
            # 如果字符串通过code_to_symbol转换为Symbol对象类型
            symbol_obj = code_to_symbol(target_symbol)
        except:
            return None
    if symbol_obj is not None:
        # 返回市场类型
        return symbol_obj.market
    return None


def market_train_test_split(n_folds, market_symbols, market=None):
    """
    切割训练集与测试集，本地训练化保存，只返回训练集
    :param n_folds: 切割比例，透传KFold中使用的参数
    :param market_symbols: 待切分的总market_symbols
    :param market: 待切分的市场，eg：EMarketTargetType.E_MARKET_TARGET_US
    :return: 返回训练集symbols数据
    """
    market_symbols, test_symbols = _do_market_train_test_split(n_folds=n_folds, market_symbols=market_symbols)

    if market is None:
        # None则服从ABuEnv.g_market_target市场设置
        market = ABuEnv.g_market_target
    market_name = market.value
    # 匹配对应的市场组成市场名称
    last_path_train = '{}_{}'.format(K_MARKET_TRAIN_FN_BASE, market_name)
    last_path_test = '{}_{}'.format(K_MARKET_TEST_FN_BASE, market_name)

    tt = datetime.datetime.now().strftime('%Y_%m_%d_%H_%M_%S_%f')
    if ABuFileUtil.file_exist(last_path_train):
        # 之前如果存在训练集切割数据，使用当前时间做记号，另保存起来
        store_train_name = '{}_{}'.format(last_path_train, tt)
        os.rename(last_path_train, store_train_name)
    if ABuFileUtil.file_exist(last_path_test):
        # 之前如果存在测试集切割数据，使用当前时间做记号，另保存起来
        store_test_name = '{}_{}'.format(last_path_test, tt)
        os.rename(last_path_test, store_test_name)

    # 本地序列化测试集数据
    ABuFileUtil.dump_pickle(test_symbols, last_path_test)
    # 本地序列化训练集数据
    ABuFileUtil.dump_pickle(market_symbols, last_path_train)
    # 只返回训练集symbols数据
    return market_symbols


def market_last_split_test(market=None):
    """
    使用最后一次切割好的测试集symbols数据
    :param market: 待获取测试集市场，eg：EMarketTargetType.E_MARKET_TARGET_US
    :return: 最后一次切割好的测试集symbols数据
    """
    if market is None:
        # None则服从ABuEnv.g_market_target市场设置
        market = ABuEnv.g_market_target
    market_name = market.value
    # 匹配对应的市场组成市场名称
    last_path_test = '{}_{}'.format(K_MARKET_TEST_FN_BASE, market_name)

    if not ABuFileUtil.file_exist(last_path_test):
        raise RuntimeError('g_enable_last_split_test not fileExist(fn)!')
    market_symbols = ABuFileUtil.load_pickle(last_path_test)
    return market_symbols


def market_last_split_train(market=None):
    """
    使用最后一次切割好的训练集symbols数据
    :param market: 待获取测试集市场，eg：EMarketTargetType.E_MARKET_TARGET_US
    :return: 最后一次切割好的训练集symbols数据
    """
    if market is None:
        # None则服从ABuEnv.g_market_target市场设置
        market = ABuEnv.g_market_target
    market_name = market.value
    # 匹配对应的市场组成市场名称
    last_path_train = '{}_{}'.format(K_MARKET_TRAIN_FN_BASE, market_name)

    if not ABuFileUtil.file_exist(last_path_train):
        raise RuntimeError('g_enable_last_split_train not ZCommonUtil.fileExist(fn)!')
    market_symbols = ABuFileUtil.load_pickle(last_path_train)
    return market_symbols


def _do_market_train_test_split(n_folds=10, market_symbols=None, shuffle=True, market=None):
    """
    分割市场训练集与测试集
    :param market_symbols，备选symbols序列，如果None, 则从参数market选择全市场symbol做为备选
    :param n_folds: 切割比例，KFold中使用的参数
    :param shuffle: 是否将原始序列打乱，默认True
    :param market: 默认None，如None则服从ABuEnv.g_market_target市场设置
    :return:
    """
    if market is None:
        # 如None则服从ABuEnv.g_market_target市场设置
        market = ABuEnv.g_market_target

    if market_symbols is None:
        # 如果None, 则从参数market选择全市场symbol做为备选
        market_symbols = all_symbol(market=market, ss=True)

    if not isinstance(market_symbols, np.ndarray):
        # 使用np.array包装
        market_symbols = np.array(market_symbols)

    n_folds = n_folds if len(market_symbols) > n_folds else len(market_symbols)
    # 使用KFold对market_symbols进行训练集，测试集拆分
    kf = KFold(len(market_symbols), n_folds=n_folds, shuffle=shuffle)
    train = None
    test = None
    for train_index, test_index in kf:
        train, test = market_symbols[train_index], market_symbols[test_index]
        # 暂时只保留一组就够了，多层级切割回测还未进行代码迁移
        break
    if train is not None and test is not None:
        # noinspection PyUnresolvedReferences
        return train.tolist(), test.tolist()
    return list(), list()


def is_in_sand_box(symbol):
    """判定symbol是否在沙盒数据支持里面"""
    cs = code_to_symbol(symbol, rs=False)
    if cs is None:
        return False
    if cs.is_futures() or cs.is_tc():
        # 沙盒数据支持完整期货和电子货币市场
        return True
    if symbol in K_SAND_BOX_CN \
            or symbol in K_SAND_BOX_US \
            or symbol in K_SAND_BOX_HK:
        # A股，美股，港股沙盒数据需要在沙盒数据序列中
        return True
    return False


"""＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊Deprecated 旧数据格式文件＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊＊"""
_rom_dir = ABuEnv.g_project_rom_data_dir
_stock_code_cn = os.path.join(_rom_dir, 'stock_code_cn.txt')
_stock_code_us = os.path.join(_rom_dir, 'stock_code_us.txt')
_stock_code_hk = os.path.join(_rom_dir, 'stock_code_hk.txt')


@AbuDeprecated()
def _parse_code(line, index):
    if ABuEnv.g_is_py3:
        line = line.decode()
    _slice = line.split('|')
    market = _slice[0]
    is_index = _slice[1] == 4
    code = _slice[2]
    if market == 'us':
        m_type = AbuSymbolUS().query_symbol_sub_market(code)
    else:
        m_type = EMarketSubType(market)
    if m_type is not None and not (not index and is_index):
        return Symbol(ABuEnv.g_market_target, m_type, code)


@AbuDeprecated('only read old symbol, miss update!!!')
def _all_us_symbol_deprecated(index=False):
    """
    获取美股全市场股票代码
    :param index: 是否包含指数
    :return:
    """
    with open(_stock_code_us, 'rb') as us_f:
        return list(filter(lambda x: x is not None, [_parse_code(line, index) for line in us_f.readlines()]))


@AbuDeprecated('only read old symbol, miss update!!!')
def _all_cn_symbol_deprecated(index=False):
    """
    获取A股全市场股票代码
    :param index: 是否包含指数
    :return:
    """
    with open(_stock_code_cn, 'rb') as cn_f:
        return list(filter(lambda x: x is not None, [_parse_code(line, index) for line in cn_f.readlines()]))


@AbuDeprecated('only read old symbol, miss update!!!')
def _all_hk_symbol_deprecated(index=False):
    """
    获取港股全市场股票代码
    :param index: 是否包含指数
    :return:
    """
    with open(_stock_code_hk, 'rb') as hk_f:
        return list(filter(lambda x: x is not None, [_parse_code(line, index) for line in hk_f.readlines()]))



from __future__ import absolute_import

from .ABuGridSearch import ParameterGrid, GridSearch
from .ABuMetricsBase import AbuMetricsBase, MetricsDemo
from .ABuMetricsFutures import AbuMetricsFutures
from .ABuMetricsTC import AbuMetricsTC
from .ABuMetricsScore import AbuBaseScorer, WrsmScorer, AbuScoreTuple, make_scorer

from . import ABuGridHelper
from . import ABuMetrics as metrics

__all__ = [
    'ParameterGrid',
    'GridSearch',
    'AbuMetricsBase',
    'AbuMetricsFutures',
    'AbuMetricsTC',
    'MetricsDemo',
    'AbuBaseScorer',
    'WrsmScorer',
    'make_scorer',
    'ABuGridHelper',
    'metrics']



# coding=utf-8
"""
    股票类型的symbol模块，a股，美股，港股
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os

import pandas as pd

from ..CoreBu.ABuFixes import six
from ..CoreBu.ABuBase import FreezeAttrMixin
from ..CoreBu import ABuEnv
from ..CoreBu.ABuEnv import EMarketTargetType, EMarketSubType
from ..UtilBu.ABuDTUtil import singleton
from ..UtilBu.ABuStrUtil import digit_str
from ..MarketBu.ABuSymbol import Symbol, code_to_symbol
from ..CrawlBu.ABuXqConsts import columns_map

__author__ = '阿布'
__weixin__ = 'abu_quant'

_rom_dir = ABuEnv.g_project_rom_data_dir
"""a股symbol，文件定期重新爬取，更新"""
_stock_code_cn = os.path.join(_rom_dir, 'stock_code_CN.csv')
"""美股symbol，文件定期重新爬取，更新"""
_stock_code_us = os.path.join(_rom_dir, 'stock_code_US.csv')
"""港股symbol，文件定期重新爬取，更新"""
_stock_code_hk = os.path.join(_rom_dir, 'stock_code_HK.csv')


class AbuStockBaseWrap(object):
    """做为类装饰器封装替换init 解析csv symbol数据操作，装饰替换init"""

    def __call__(self, cls):
        if isinstance(cls, six.class_types):
            # 只做为类装饰器使用，拿出原始的__init__
            init = cls.__init__

            def wrapped(*args, **kwargs):
                warp_self = args[0]
                warp_self.df = None

                init(*args, **kwargs)

                # 剔除重复数据
                warp_self.df.drop_duplicates(inplace=True)

                warp_self.df['industry_factorize'], industry_factorize_name = \
                    pd.factorize(warp_self.df.industry)
                # 用Series包装一下离散后的行业信息，以便方便对应行业索引
                warp_self.industry_factorize_name_series = pd.Series(industry_factorize_name)

                # 将映射中key和value进行互换，columns_map中中文的key和英文的value（详ABuXqConsts），即形成本地语言列名
                local_columns_map = {columns_map[col_key]: col_key for col_key in columns_map}
                # show_df中列名使用本地语言展示
                warp_self.show_df = warp_self.df.rename(columns=local_columns_map, inplace=False)
                # 冻结接口，只读
                # noinspection PyProtectedMember
                warp_self._freeze()

            # 使用wrapped替换初始的__init__
            cls.__init__ = wrapped
            wrapped.__name__ = '__init__'
            # 记录原始init的方法为deprecated_original
            wrapped.deprecated_original = init
            return cls
        else:
            raise TypeError('AbuStockBaseWrap just for class warp')


# noinspection PyUnresolvedReferences
class AbuSymbolStockBase(FreezeAttrMixin):
    """股票类型的symbol抽象基类"""

    def __str__(self):
        """打印对象显示：df.info， df.describe"""
        return 'info:\n{}\ndescribe:\n{}'.format(self.df.info(),
                                                 self.df.describe())

    __repr__ = __str__

    def __len__(self):
        """对象长度：df.shape[0]，即行数"""
        return self.df.shape[0]

    def __setitem__(self, key, value):
        """索引设置：对外抛出错误， 即不准许外部设置"""
        raise AttributeError("AbuFuturesCn set value!!!")

    def query_industry_symbols(self, query_symbol, local_df=True):
        """
        为ABuIndustries模块，提供查询股票所在的行业industry_df子集
        :param query_symbol: symbol str对象
        :param local_df: 是否基于show_df返回行业df
        :return: 查询query_symbol所在的行业对象，pd.DataFrame对象
        """
        industry_df = None
        # 使用in查询self，即子类都需要实现自己的__contains__
        if query_symbol in self:
            # 忽略一个问题，如果只使用000001不带子市场标识去查询，结果只取第一个，准确查询需要完整标示
            factorize = self[query_symbol]['industry_factorize'].values[0]
            # 参数local_df决定行业industry_df子集基于本地语言的show_df还是df，使用show_df更加方便外面查阅对应内容意义
            base_df = self.show_df if local_df else self.df
            # 切取行业factorize相同的子集
            industry_df = base_df[base_df['industry_factorize'] == factorize]
        return industry_df

    def query_industry_factorize(self, factorize, local_df=True):
        """
        为ABuIndustries模块，提行业的factorize值查询industry_df子集
        :param factorize: int
        :param local_df: 是否基于show_df返回行业df
        :return: 通过factorize值查询industry_df子集，pd.DataFrame对象
        """
        base_df = self.show_df if local_df else self.df
        # 切取行业factorize相同的子集
        industry_df = base_df[base_df['industry_factorize'] == factorize]
        return industry_df

    def all_symbol(self, index=False):
        """
        子类需要实现，获取市场中所有股票symbol str对象序列，即全市场symbol序列
        :param index: 是否需要返回大盘symbol
        :return: 全市场symbol序列
        """
        raise NotImplementedError('NotImplementedError AbuSymbolStockBase all_symbol!!!')

    def query_symbol_sub_market(self, *args, **kwargs):
        """
        子类需要实现，查询股票所在的子市场，即交易所信息
        :return: 返回EMarketSubType.value值，即子市场（交易所）字符串对象
        """
        raise NotImplementedError('NotImplementedError AbuSymbolStockBase query_symbol_sub_market!!!')

    def symbol_func(self, df):
        """
        子类需要实现，通过df组装支持ABuSymbolPd.make_kl_df使用的symbol
        :param df: pd.DataFrame对象
        :return: 支持ABuSymbolPd.make_kl_df使用的symbol序列
        """
        raise NotImplementedError('NotImplementedError AbuSymbolStockBase symbol_func!!!')


@singleton
@AbuStockBaseWrap()
class AbuSymbolCN(AbuSymbolStockBase):
    """a股symbol类，singleton"""

    def __init__(self):
        """被AbuStockBaseWrap替换__init__，即只需读取a股数据到self.df 后续在类装饰器完成"""
        self.df = pd.read_csv(_stock_code_cn, index_col=0, dtype=str)

    def __contains__(self, item):
        """成员测试：是否item在self.df.symbol.values中"""
        return digit_str(item) in self.df.symbol.values

    def __getitem__(self, key):
        """
            索引获取：两种模式索引获取：
            1. 参数key为df的columns名称，返回self.df[key]，即get df的列
            2. 参数key为股票代码名称，标准化后查询，self.df[self.df.symbol == key]，即get df的行
        """

        if key in self.df.columns:
            # 参数key为df的columns名称，get df的行
            return self.df[key]

        if len(key) > 2:
            head = key[:2].upper()
            if head.isalpha():
                # 头两位是字面，即认为是exchange信息，直接截取df_filter
                df_filter = self.df[self.df['exchange'] == head]
                if not df_filter.empty:
                    if key[2:] in df_filter.symbol.values:
                        # get df的行信息，即对应股票的所有信息
                        return df_filter[df_filter.symbol == key[2:]]
            else:
                if key in self.df.symbol.values:
                    # get df的行
                    return self.df[self.df.symbol == key]

    def symbol_func(self, df):
        """
        通过df组装支持ABuSymbolPd.make_kl_df使用的symbol，
        使用df['exchange'].map(lambda exchange: exchange.lower()) + df['symbol']
        :param df: pd.DataFrame对象
        :return: 支持ABuSymbolPd.make_kl_df使用的symbol序列
        """
        df_symbol = df['exchange'].map(lambda exchange: exchange.lower()) + df['symbol']
        return df_symbol.tolist()

    def all_symbol(self, index=False):
        """
        获取a股市场中所有股票symbol str对象序列，即a股全市场symbol序列
        :param index: 是否需要返回a股大盘symbol
        :return: a股全市场symbol序列
        """

        # 过滤df中的A股指数symbol
        a_index = self.industry_factorize_name_series[self.industry_factorize_name_series == 'A股指数'].index.values[0]
        df_filter = self.df[self.df['industry_factorize'] != a_index]
        # 通过symbol_func转换为外部可直接使用ABuSymbolPd.make_kl_df请求的symbol序列
        all_symbol = self.symbol_func(df_filter)
        if index:
            # 需要返回大盘symbol
            all_symbol.extend(['{}{}'.format(EMarketSubType.SH.value, symbol) for symbol in Symbol.SH_INDEX])
            all_symbol.extend(['{}{}'.format(EMarketSubType.SZ.value, symbol) for symbol in Symbol.SZ_INDEX])
        return all_symbol

    def query_symbol_sub_market(self, code, default=EMarketSubType.SH.value):
        """
        查询股票所在的子市场，即交易所信息, A股市场默认返回上证交易所
        :return: 返回EMarketSubType.value值，即子市场（交易所）字符串对象
        """

        if code in self:
            # 忽略一个问题，如果只使用000001不带子市场标识去查询，结果只取第一个，准确查询需要完整标示
            return self[code].market.values[0].lower()

        # 如果没查到如果首symbol为6，9为判定为sh
        if code[:1] in ['6', '9']:
            return EMarketSubType.SH.value
        # 如果没查到如果首symbol为2，3为判定为sz
        elif code[:1] in ['2', '3']:
            return EMarketSubType.SZ.value
        return default


@singleton
@AbuStockBaseWrap()
class AbuSymbolUS(AbuSymbolStockBase):
    """美股symbol类，singleton"""

    """针对历史不适合做回测，对回测结果有误导影响的symbol, 即可能会产生几千倍，几百倍收益的不要参加回测"""
    s_unusual_symbol = ['usACV', 'usAMPH', 'usCBX', 'usDCIX', 'usDM', 'usEPE', 'usFPL', 'usFUEL', 'usGDI', 'usHCC',
                        'usKBSF', 'usKEG', 'usKMI', 'usLMCA', 'usLTM', 'usLUX', 'usMBRX', 'usMPG', 'usOPXAW', 'usORN',
                        'usPJT', 'usPTIE', 'usSAB', 'usSPR', 'usSR', 'usTGEN', 'usTNXP', 'usVBIV', 'usWMGIZ',
                        'usXGTIW', 'usMBRX']

    def __init__(self):
        """被AbuStockBaseWrap替换__init__，即只需读取美股数据到self.df 后续在类装饰器完成"""
        self.df = pd.read_csv(_stock_code_us, index_col=0, dtype=str)

    def __contains__(self, item):
        """成员测试：是否item或item[2:]在self.df.symbol.values中"""
        return item in self.df.symbol.values or (len(item) > 2 and item[2:] in self.df.symbol.values)

    def __getitem__(self, key):
        """
            索引获取：两种模式索引获取：
            1. 参数key为df的columns名称，返回self.df[key]，即get df的列
            2. 参数key为股票代码名称，标准化后查询，self.df[self.df.symbol == key]，即get df的行
        """

        if key in self.df.columns:
            # 参数key为df的columns名称，返回self.df[key]，即get df的列
            return self.df[key]

        # get df的行, 即对于股票的详细信息
        if key in self.df.symbol.values:
            return self.df[self.df.symbol == key]
        if len(key) > 2 and key[2:] in self.df.symbol.values:
            return self.df[self.df.symbol == key[2:]]

    def symbol_func(self, df):
        """
        通过df组装支持ABuSymbolPd.make_kl_df使用的symbol，使用('us' + df['symbol']).tolist()
        :param df: pd.DataFrame对象
        :return: 支持ABuSymbolPd.make_kl_df使用的symbol序列
        """
        # noinspection PyUnresolvedReferences
        return (EMarketTargetType.E_MARKET_TARGET_US.value + df['symbol']).tolist()

    def all_symbol(self, index=False):
        """
        获取美股市场中所有股票symbol str对象序列，即美股全市场symbol序列
        :param index: 是否需要返回美股大盘symbol
        :return: 美股全市场symbol序列
        """

        # 过滤AMEX等大盘，etf类型, 只取NASDAQ和NYSE
        df_filter = self.df[(self.df['exchange'] == EMarketSubType.US_OQ.value) |
                            (self.df['exchange'] == EMarketSubType.US_N.value)]
        # 通过symbol_func转换为外部可直接使用ABuSymbolPd.make_kl_df请求的symbol序列
        all_symbol = self.symbol_func(df_filter)
        all_symbol = list(set(all_symbol) - set(AbuSymbolUS.s_unusual_symbol))
        if index:
            # 需要返回大盘symbol
            all_symbol.extend(['{}{}'.format(EMarketTargetType.E_MARKET_TARGET_US.value, symbol)
                               for symbol in Symbol.US_INDEX])
        return all_symbol

    def query_symbol_sub_market(self, code, default=EMarketSubType.US_N.value):
        """
        查询股票所在的子市场，即交易所信息, 美股市场默认返回纽交所
        :return: 返回EMarketSubType.value值，即子市场（交易所）字符串对象
        """

        if code in self:
            return self[code].exchange.values[0].upper()
        return default


@singleton
@AbuStockBaseWrap()
class AbuSymbolHK(AbuSymbolStockBase):
    """港股symbol类，singleton"""

    def __init__(self):
        """被AbuStockBaseWrap替换__init__，即只需读取港股数据到self.df 后续在类装饰器完成"""
        self.df = pd.read_csv(_stock_code_hk, index_col=0, dtype=str)

    def __contains__(self, item):
        """成员测试：是否item在self.df.symbol.values中"""
        return digit_str(item) in self.df.symbol.values

    def __getitem__(self, key):
        """
            索引获取：两种模式索引获取：
            1. 参数key为df的columns名称，返回self.df[key]，即get df的列
            2. 参数key为股票代码名称，标准化后查询，self.df[self.df.symbol == key]，即get df的行
        """

        # 参数key为df的columns名称，返回self.df[key]
        if key in self.df.columns:
            return self.df[key]

        # 参数key为股票代码名称，标准化后查询
        ds = digit_str(key)
        if ds in self.df.symbol.values:
            return self.df[self.df.symbol == ds]

    def symbol_func(self, df):
        """
        通过df组装支持ABuSymbolPd.make_kl_df使用的symbol，使用('hk' + df['symbol']).tolist()
        :param df: pd.DataFrame对象
        :return: 支持ABuSymbolPd.make_kl_df使用的symbol序列
        """
        # noinspection PyUnresolvedReferences
        return (EMarketTargetType.E_MARKET_TARGET_HK.value + df['symbol']).tolist()

    def all_symbol(self, index=False):
        """
         获取港股市场中所有股票symbol str对象序列，即港股全市场symbol序列
         :param index: 是否需要返回港股大盘symbol
         :return: 港股全市场symbol序列
         """

        # 通过symbol_func转换为外部可直接使用ABuSymbolPd.make_kl_df请求的symbol序列
        all_symbol = self.symbol_func(self.df)
        if index:
            # 需要返回大盘symbol
            all_symbol.extend(['{}{}'.format(EMarketTargetType.E_MARKET_TARGET_HK.value, symbol)
                               for symbol in Symbol.HK_INDEX])
        return all_symbol

    # noinspection PyUnusedLocal
    def query_symbol_sub_market(self, code, default=EMarketSubType.HK.value):
        """
        查询股票所在的子市场，即交易所信息, 港股市场默认返回hk
        :return: 返回EMarketSubType.value值，即子市场（交易所）字符串对象
        """
        return default


def query_stock_info(symbol):
    """
    通过将symbol code转换为Symbol对象查询对应的市场，构造对应的市场对象，
    仅支持股票类型symbol
    :param symbol: eg：usTSLA
    :return: 一行数据的pd.DataFrame对象
    """
    if isinstance(symbol, six.string_types):
        symbol = code_to_symbol(symbol)

    if symbol.is_a_stock():
        sn = AbuSymbolCN()
    elif symbol.is_hk_stock():
        sn = AbuSymbolHK()
    elif symbol.is_us_stock():
        sn = AbuSymbolUS()
    else:
        print('query_symbol_info just suit sz, sh, us, hk!')
        return
    # 直接使用类的__getitem__方法
    return sn[symbol.symbol_code]



# -*- encoding:utf-8 -*-
"""
    辅助进度显示模块，多进程，单进程
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import sys
import os
import datetime
import time

from IPython.display import clear_output
from IPython.display import display
from ipywidgets import FloatProgress, Text, Box

from ..CoreBu import ABuEnv
from ..UtilBu.ABuDTUtil import warnings_filter, catch_error
from ..UtilBu import ABuFileUtil, ABuOsUtil
from ..CoreBu.ABuParallel import run_in_subprocess, run_in_thread

__author__ = '阿布'
__weixin__ = 'abu_quant'


def do_clear_output(wait=False):
    """
    模块方法，clear所有的输出，内部针对notebook和命令行输出做区分
    :param wait: 是否同步执行clear操作，透传给IPython.display.clear_output
    """
    if ABuEnv.g_is_ipython:
        # notebook clear
        clear_output(wait=wait)
    else:
        # cmd clear
        cmd = 'clear' if ABuEnv.g_is_mac_os else 'cls'
        os.system(cmd)
        # pass


class UIProgress(object):
    """多进程socket通信下的进度显示类"""

    def __init__(self, a_pid):
        """通过进程pid初始化ui组件"""
        self.progress_widget = FloatProgress(value=0, min=0, max=100)
        self.text_widget = Text('pid={} begin work'.format(a_pid))
        # 通过box容器都放到一个里面
        self.progress_box = Box([self.text_widget, self.progress_widget])
        display(self.progress_box)

    def update(self, p_progress, p_progress_text):
        """进度条更新以及对应文字更新"""
        self.progress_widget.value = p_progress
        self.text_widget.value = p_progress_text

    def close(self):
        """关闭ui显示"""
        self.progress_box.close()


"""多进程下进度条通信socket文件基础名字"""
K_SOCKET_FN_BASE = os.path.join(ABuEnv.g_project_cache_dir, 'abu_socket_progress')
"""多进程下进度条通信socket文件最终名字，这里子进程可以获取g_socket_fn是通过ABuEnvProcess拷贝了主进程全局信息"""
g_socket_fn = None
"""多进程下进度是否显示ui进度，只针对进程间通信类型的进度，有些太频繁的进度显示可以选择关闭"""
g_show_ui_progress = True
"""主进程下用来存贮子进程传递子进程pid为key，进度条对象UIProgress为value"""
ui_progress_dict = {}


def _socket_cmd_handle(socket_cmd):
    """主进程中处理子进程传递的进度条处理信息：创建，进度更新，销毁"""
    socket_cmd = socket_cmd.strip()
    socket_cmd = socket_cmd.strip('\x00')
    socket_cmd = socket_cmd.strip('\0')
    cmd_split = socket_cmd.split('|')
    if len(cmd_split) == 3 and cmd_split[0] in ui_progress_dict:
        # 3个字段的0是：pid
        pid = cmd_split[0]
        # 3个字段的1是：进度，转换float
        progress = float(cmd_split[1])
        # 3个字段的2是：进度文字显示
        progress_text = cmd_split[2]
        # 找到字典中的UIProgress对象开始update
        ui_progress_dict[pid].update(progress, progress_text)
    elif len(cmd_split) == 2 and cmd_split[0] in ui_progress_dict and cmd_split[1] == 'close':
        # 2个字段的0是：pid
        pid = cmd_split[0]
        # 将字典中的UIProgress对象pop出来，后执行close
        pop_progress = ui_progress_dict.pop(pid, None)
        if pop_progress is not None:
            pop_progress.close()
    elif len(cmd_split) == 2 and cmd_split[1] == 'init':
        pid = cmd_split[0]
        if pid in ui_progress_dict:
            ui_progress_dict.pop(pid)
        # 创建的进度条以pid为key放到缓存字典中
        ui_progress_dict[pid] = UIProgress(pid)


# 不管ui进度条有什么问题，也不能影响任务主进程主任务的进度执行，反正有文字进度会始终显示
@catch_error(log=False)
def ui_progress_socket_work():
    """主进程下的子线程函数：子进程传递的进度条处理信息：创建，进度更新，销毁"""
    global g_socket_fn

    # 不管是共享内存实现还是socket都通过当前时间＋pid确定唯一文件名称
    tt = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S_%f")
    if ABuEnv.g_is_mac_os:
        # bsd socket
        g_socket_fn = '{}_{}_{}'.format(K_SOCKET_FN_BASE, os.getpid(), tt)
        if ABuFileUtil.file_exist(g_socket_fn):
            # 如果socket文件存在，删除
            ABuFileUtil.del_file(g_socket_fn)
    else:
        # windows全局共享内存通过pid＋时间确定
        g_socket_fn = 'ABU_PROGRESS_{}_{}'.format(os.getpid(), tt)

    # socket_bind_recv不管是windows还是mac都在这里进行while True阻塞子线程一直运行
    ABuOsUtil.socket_bind_recv(g_socket_fn, _socket_cmd_handle)


# 不管ui进度条有什么问题，也不能影响任务主进程主任务工作的进度执行
@catch_error(log=False)
def check_process_is_dead():
    """主进程下的子线程函数：检测ui_progress_dict中的进程pid是否仍然活着，如果死了，从字典中清除，close ui"""
    while True:
        # 低优先级任务，1分钟执行1次
        time.sleep(60)
        do_check_process_is_dead()


def do_check_process_is_dead():
    """执行检测ui_progress_dict中的进程pid是否仍然活着，如果死了，从字典中清除，close ui"""
    import psutil
    # 获取活着的所有pid序列
    living = psutil.pids()
    clear_arr = list()
    for progress_pid in ui_progress_dict:
        # 需要临时转换一次int，living中进程序列是int
        if int(progress_pid) not in living:
            # 字典中记录的pid如果不在活着的序列中，清除
            clear_arr.append(progress_pid)
    for clear_pid in clear_arr:
        if clear_pid in ui_progress_dict:
            pop_progress = ui_progress_dict.pop(clear_pid, None)
            if pop_progress is not None:
                pop_progress.close()

if g_show_ui_progress and ABuEnv.g_main_pid == os.getpid() and ABuEnv.g_is_ipython:
    # 如果是主进程执行进行子线程函数ui_progress_socket_work：处理子进程传递的进度条处理信息：创建，进度更新，销毁
    run_in_thread(ui_progress_socket_work)
    # 如果是主进程执行进行子线程函数check_process_is_dead：检测ui_progress_dict中的进程是否仍然活着，从字典中清除，close ui
    run_in_thread(check_process_is_dead)


class AbuMulPidProgress(object):
    """多进程进度显示控制类"""

    def __init__(self, total, label):
        """
        外部使用eg：
        with AbuMulPidProgress(len(self.choice_symbols), 'pick stocks complete') as progress:
            for epoch, target_symbol in enumerate(self.choice_symbols):
                progress.show(epoch + 1)

        :param total: 总任务数量
        :param label: 进度显示label
        """
        self._total = total
        self._label = label
        self.epoch = 0
        self.display_step = 1
        self.progress_widget = None
        self.text_widget = None
        self.progress_box = None

    # 不管ui进度条有什么问题，也不能影响任务工作的进度执行，反正有文字进度会始终显示
    @catch_error(log=False)
    def init_ui_progress(self):
        """初始化ui进度条"""

        if not ABuEnv.g_is_ipython or self._total < 2:
            return

        if ABuEnv.g_main_pid == os.getpid():
            # 如果是在主进程下显示那就直接来
            self.progress_widget = FloatProgress(value=0, min=0, max=100)
            self.text_widget = Text('pid={} begin work'.format(os.getpid()))
            self.progress_box = Box([self.text_widget, self.progress_widget])
            display(self.progress_box)
        else:
            if g_show_ui_progress and g_socket_fn is not None:
                # 子进程下通过socket通信将pid给到主进程，主进程创建ui进度条
                ABuOsUtil.socket_send_msg(g_socket_fn, '{}|init'.format(os.getpid()))

    # 不管ui进度条有什么问题，也不能影响任务工作的进度执行，反正有文字进度会始终显示
    @catch_error(log=False)
    def update_ui_progress(self, ps, ps_text):
        """更新文字进度条"""

        if not ABuEnv.g_is_ipython or self._total < 2:
            return

        if ABuEnv.g_main_pid == os.getpid():
            # 如果是在主进程下显示那就直接来
            if self.progress_widget is not None:
                self.progress_widget.value = ps
            if self.text_widget is not None:
                self.text_widget.value = ps_text
        else:
            if g_show_ui_progress and g_socket_fn is not None:
                # 子进程下通过socket通信将pid给到主进程，主进程通过pid查找对应的进度条对象后更新进度
                ABuOsUtil.socket_send_msg(g_socket_fn, '{}|{}|{}'.format(os.getpid(), ps, ps_text))

    # 不管ui进度条有什么问题，也不能影响任务工作的进度执行，反正有文字进度会始终显示
    @catch_error(log=False)
    def close_ui_progress(self):
        """关闭ui进度条显示"""

        if not ABuEnv.g_is_ipython or self._total < 2:
            return

        if ABuEnv.g_main_pid == os.getpid():
            # 如果是在主进程下显示那就直接来
            if self.progress_box is not None:
                self.progress_box.close()
        else:
            if g_show_ui_progress and g_socket_fn is not None:
                # 子进程下通过socket通信将pid给到主进程，主进程通过pid查找对应的进度条对象后关闭对象，且弹出
                ABuOsUtil.socket_send_msg(g_socket_fn, '{}|close'.format(os.getpid()))

    def __enter__(self):
        """
        以上下文管理器类方式实现__enter__，针对self._total分配self.display_step
        """
        self.display_step = 1
        if self._total >= 5000:
            self.display_step = 50
        elif self._total >= 3000:
            self.display_step = 30
        elif self._total >= 2000:
            self.display_step = 20
        elif self._total > 1000:
            self.display_step = 10
        elif self._total >= 600:
            self.display_step = 6
        elif self._total >= 300:
            self.display_step = 3
        elif self._total >= 100:
            self.display_step = 2
        elif self._total >= 20:
            self.display_step = 2
        self.epoch = 0
        self.init_ui_progress()

        return self

    def show(self, epoch=None, clear=True):
        """
        进行进度控制显示主方法
        :param epoch: 默认None, 即使用类内部计算的迭代次数进行进度显示
        :param clear: 默认True, 子进程显示新的进度前，先do_clear_output所有输出
        :return:
        """
        self.epoch = epoch if epoch is not None else self.epoch + 1
        if self.epoch % self.display_step == 0:
            ps = round(self.epoch / self._total * 100, 2)
            ps = 100 if ps > 100 else ps
            ps_text = "pid:{} {}:{}%".format(os.getpid(), self._label, ps)
            if not ABuEnv.g_is_ipython or self._total < 2:
                if clear:
                    do_clear_output()
                print(ps_text)
            self.update_ui_progress(ps, ps_text)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        以上下文管理器类方式实现__exit__，针对在子进城中的输出显示进度进行do_clear_output扫尾工作
        """
        clear = False
        if clear:
            # clear在mac上应该打开, 由于windows某些版本浏览器wait=True会有阻塞情况，如果wait＝False, 有clear之后的风险，
            do_clear_output(wait=True)  # wait 需要同步否则会延迟clear
        else:
            print("pid:{} done!".format(os.getpid()))

        self.close_ui_progress()


class AbuBlockProgress(object):
    """主进程阻塞任务，启动子单进程任务进度显示控制类"""

    def __init__(self, label, interval=1, max_step=20):
        """
        :param label: 阻塞进度条显示的文字信息
        :param interval: 阻塞进度条显示的时间间隔
        :param max_step: 进度最大显示粒度
        """
        self.label = label
        self.interval = interval
        self.sub_process = None
        self.max_step = max_step

    def __enter__(self):
        """创建子进程做进度显示"""

        def progress_interval(interval, label):
            count = 1
            while True:
                p_str = '*^{}s'.format(int(count * 3))
                end = format('*', p_str)
                progress_str = '{}{}'.format(label, end)
                do_clear_output()
                print(progress_str)
                count += 1
                if count > self.max_step:
                    count = 1
                time.sleep(interval)

        self.sub_process = run_in_subprocess(progress_interval, self.interval, self.label)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """结束子进程，清理输出"""
        if self.sub_process is not None and self.sub_process.is_alive():
            self.sub_process.terminate()
            do_clear_output()


class AbuProgress(object):
    """单进程（主进程）进度显示控制类"""

    # 过滤DeprecationWarning: Widget._keys_default is deprecated in traitlets 4.1: use @default decorator instead.
    @warnings_filter
    def __init__(self, total, a_progress, label=None):
        """
        外部使用eg：
            progess = AbuProgress(stock_df.shape[0], 0, 'merging {}'.format(m))
            for i, symbol in enumerate(stock_df['symbol']):
                progess.show(i + 1)
        :param total: 总任务数量
        :param a_progress: 初始进度
        :param label: 进度显示label
        """
        self._total = total
        self._progress = a_progress
        self._label = label
        self.f = sys.stdout
        self.progress_widget = None

    def __enter__(self):
        """创建子进程做进度显示"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.f.write('\r')
        if self.progress_widget is not None:
            self.progress_widget.close()

    @property
    def progress(self):
        """property获取self._progress"""
        return self._progress

    @progress.setter
    def progress(self, a_progress):
        """rogress.setter设置progress"""
        if a_progress > self._total:
            self._progress = self._total
        elif a_progress < 0:
            self._progress = 0
        else:
            self._progress = a_progress

    def show(self, a_progress=None, ext='', p_format="{}:{}:{}%"):
        """
        进行进度控制显示主方法
        :param ext: 可以添加额外的显示文字，str，默认空字符串
        :param a_progress: 默认None, 即使用类内部计算的迭代次数进行进度显示
        :param p_format: 进度显示格式，默认{}: {}%，即'self._label:round(self._progress / self._total * 100, 2))%'
        """
        self.progress = a_progress if a_progress is not None else self.progress + 1
        ps = round(self._progress / self._total * 100, 2)

        if self._label is not None:
            # 如果初始化label没有就只显示ui进度
            self.f.write('\r')
            self.f.write(p_format.format(self._label, ext, ps))

        if ABuEnv.g_is_ipython:
            if self.progress_widget is None:
                self.progress_widget = FloatProgress(value=0, min=0, max=100)
                display(self.progress_widget)
            self.progress_widget.value = ps

        # 这样会出现余数结束的情况，还是尽量使用上下文管理器控制结束
        if self._progress == self._total:
            self.f.write('\r')
            if self.progress_widget is not None:
                self.progress_widget.close()



# -*- encoding:utf-8 -*-
"""
    黄金分割及比例分割示例模块
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from collections import namedtuple

import matplotlib.pyplot as plt

from ..TLineBu import ABuTLExecute

__author__ = '阿布'
__weixin__ = 'abu_quant'


def calc_golden(kl_pd, show=True):
    """
    只针对金融时间序列的收盘价格close序列，进行黄金分割及比例分割
    数值结果分析以及可视化
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param show: 是否可视化黄金分割及比例分割结果
    :return: 黄金分割及比例分割结果组成的namedtuple数值对象
    """
    kl_close = kl_pd.close

    if not hasattr(kl_pd, 'name'):
        # 金融时间序列中如果有异常的没有name信息的补上一个unknown
        kl_pd.name = 'unknown'

    # 计算视觉黄金分割
    gd_382, gd_500, gd_618 = ABuTLExecute.find_golden_point(kl_pd.index, kl_close)
    # 计算统计黄金分割
    gex_382, gex_500, gex_618 = ABuTLExecute.find_golden_point_ex(kl_pd.index, kl_close)

    # below above 382, 618确定，即382，618上下底
    below618, above618 = ABuTLExecute.below_above_gen(gd_618, gex_618)
    below382, above382 = ABuTLExecute.below_above_gen(gd_382, gex_382)

    # 再次通过比例序列percents和find_percent_point寻找对应比例的位置字典pts_dict
    percents = [0.20, 0.25, 0.30, 0.70, 0.80, 0.90, 0.95]
    pts_dict = ABuTLExecute.find_percent_point(percents, kl_close)

    # 0.20, 0.25, 0.30只找最低的，即底部只要最低的
    below200, _ = ABuTLExecute.below_above_gen(*pts_dict[0.20])
    below250, _ = ABuTLExecute.below_above_gen(*pts_dict[0.25])
    below300, _ = ABuTLExecute.below_above_gen(*pts_dict[0.30])

    # 0.70, 0.80, 0.90, 0.95只找最高的，即顶部只要最高的
    _, above700 = ABuTLExecute.below_above_gen(*pts_dict[0.70])
    _, above800 = ABuTLExecute.below_above_gen(*pts_dict[0.80])
    _, above900 = ABuTLExecute.below_above_gen(*pts_dict[0.90])
    _, above950 = ABuTLExecute.below_above_gen(*pts_dict[0.95])

    if show:
        # 开始可视化黄金分割及比例分割结果
        plt.axes([0.025, 0.025, 0.95, 0.95])
        plt.plot(kl_close)

        # 0.70, 0.80, 0.90, 0.95，lw线条粗度递减
        plt.axhline(above950, lw=3.5, color='c')
        plt.axhline(above900, lw=3.0, color='y')
        plt.axhline(above800, lw=2.5, color='k')
        plt.axhline(above700, lw=2.5, color='m')

        # 中间层的618是带，有上下底
        plt.axhline(above618, lw=2, color='r')
        plt.axhline(below618, lw=1.5, color='r')
        plt.fill_between(kl_pd.index, above618, below618,
                         alpha=0.1, color="r")
        # 中间层的382是带，有上下底
        plt.axhline(above382, lw=1.5, color='g')
        plt.axhline(below382, lw=2, color='g')
        plt.fill_between(kl_pd.index, above382, below382,
                         alpha=0.1, color="g")

        # 0.20, 0.25, 0.30 lw线条粗度递曾
        plt.axhline(below300, lw=2.5, color='k')
        plt.axhline(below250, lw=3.0, color='y')
        plt.axhline(below200, lw=3.5, color='c')

        _ = plt.setp(plt.gca().get_xticklabels(), rotation=30)
        plt.legend([kl_pd.name, 'above950', 'above900', 'above800', 'above700', 'above618', 'below618',
                    'above382', 'below382', 'below300', 'below250', 'below200'], bbox_to_anchor=(1.05, 1), loc=2,
                   borderaxespad=0.)
        plt.title('between golden')
        plt.show()

    return namedtuple('golden', ['g382', 'gex382', 'g500', 'gex500', 'g618',
                                 'gex618', 'above618', 'below618', 'above382', 'below382',
                                 'above950', 'above900', 'above800', 'above700', 'below300', 'below250', 'below200'])(
        gd_382, gex_382,
        gd_500, gex_500, gd_618, gex_618, above618, below618, above382, below382,
        above950, above900, above800, above700, below300, below250, below200)



# -*- encoding:utf-8 -*-
"""示例ump价格特征模块"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from ..TradeBu.ABuMLFeature import AbuFeaturePrice
from ..MLBu.ABuMLPd import AbuMLPd
from .ABuUmpBase import ump_main_make_xy, BuyUmpMixin
from .ABuUmpMainBase import AbuUmpMainBase

__author__ = '阿布'
__weixin__ = 'abu_quant'


class AbuUmpMainPrice(AbuUmpMainBase, BuyUmpMixin):
    """主裁价格特征类，AbuUmpMainBase子类，混入BuyUmpMixin，做为买入ump类"""

    class UmpPriceFiter(AbuMLPd):
        """
            内部类，只需要实现make_xy，且使用ump_main_make_xy装饰

            在__init__中通过：
                    self.fiter_cls = self.get_fiter_class()
                    self.fiter = self.fiter_cls(orders_pd=self.orders_pd, **kwarg)
            构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
        """

        @ump_main_make_xy
        def make_xy(self, **kwarg):
            """
            make_xy通过装饰器ump_main_make_xy进行二次包装
            这里只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df
            :param kwarg: ump_main_make_xy装饰器中使用kwarg
                          kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
                          kwargs['scaler']    做为可选关键字参数：控制在make_xy中返回的特征矩阵数据是否进行标准化处理

            :return: self.order_has_ret中使用filter选取特征列形成price_df
                     ump_main_make_xy装饰器在make_xy返回price_df后做转换matrix，形成x，y等工作
            """

            # regex='result|buy_price_rank120|buy_price_rank90|buy_price_rank60|buy_price_rank252'
            regex = 'result|{}'.format('|'.join(AbuFeaturePrice().get_feature_ump_keys(ump_cls=AbuUmpMainPrice)))
            # noinspection PyUnresolvedReferences
            price_df = self.order_has_ret.filter(regex=regex)
            """
                eg: price_df
                            result  buy_price_rank120  buy_price_rank90  buy_price_rank60  \
                2014-09-24       0              1.000             1.000             1.000
                2014-10-24       0              1.000             1.000             1.000
                2014-10-29       1              1.000             1.000             1.000
                2014-10-29       1              0.925             0.900             0.883
                2014-10-29       1              0.962             0.950             0.925
                2014-10-29       1              1.000             1.000             1.000
                2014-11-03       1              1.000             1.000             1.000
                2014-11-11       1              0.954             0.939             0.992
                2014-11-12       0              0.475             0.522             0.783
                2014-11-26       0              0.642             0.733             1.000
                ...            ...                ...               ...               ...
                2016-03-14       1              0.617             0.500             0.750
                2016-03-14       0              0.683             0.589             0.850
                2016-03-30       0              0.658             0.667             1.000
                2016-04-04       1              0.400             0.511             0.767
                2016-04-13       0              0.567             0.722             1.000
                2016-04-14       0              0.875             0.878             0.967
                2016-04-15       0              0.775             0.733             1.000
                2016-04-15       0              0.775             0.733             1.000
                2016-04-29       1              1.000             1.000             1.000
                2016-04-29       1              1.000             1.000             1.000

                            buy_price_rank252
                2014-09-24              1.000
                2014-10-24              1.000
                2014-10-29              1.000
                2014-10-29              0.750
                2014-10-29              0.982
                2014-10-29              1.000
                2014-11-03              1.000
                2014-11-11              0.808
                2014-11-12              0.560
                2014-11-26              0.762
                ...                       ...
                2016-03-14              0.444
                2016-03-14              0.623
                2016-03-30              0.536
                2016-04-04              0.190
                2016-04-13              0.270
                2016-04-14              0.940
                2016-04-15              0.631
                2016-04-15              0.631
                2016-04-29              1.000
                2016-04-29              1.000
            """
            return price_df

    def get_predict_col(self):
        """
        主裁价格特征keys：['buy_price_rank120', 'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252']
        :return: ['buy_price_rank120', 'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252']
        """
        return AbuFeaturePrice().get_feature_ump_keys(ump_cls=AbuUmpMainPrice)

    def get_fiter_class(self):
        """
        主裁价格特征返回的AbuMLPd子类：AbuUmpMainPrice.UmpPriceFiter
        :return: AbuUmpMainPrice.UmpPriceFiter
        """
        return AbuUmpMainPrice.UmpPriceFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'price_main'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见ABuUmpManager中extend_ump_block方法
        """
        return 'price_main'



# -*- encoding:utf-8 -*-
"""操作系统工具函数整合模块"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import logging

from ..CoreBu import ABuEnv
from ..UtilBu.ABuDTUtil import catch_error

log_func = logging.info if ABuEnv.g_is_ipython else print


@catch_error()
def show_msg(title, msg, log=True):
    """
    统一平台弹窗信息提示，被catch_error装饰，即不应该被提示中断程序，
    特别长任务的情况下
    :param title: 弹窗信息标题
    :param msg: 弹窗信息内容
    :param log: 是否通过logging.info打印信息
    :return:
    """
    # 由于catch_error忽略错误，所有主要信息还是先打印
    if log:
        log_func('{}\n{}'.format(title, msg))
    if ABuEnv.g_is_mac_os:
        from ..UtilBu.ABuMacUtil import show_msg as do_show_msg
    else:
        from ..UtilBu.ABuWinUtil import show_msg as do_show_msg
    do_show_msg(title, msg)


def socket_bind_recv(socket_fn, cmd_handler):
    """
    进程间socket或者共享内存通信，接受消息，处理消息，外层应处理catch_error
    :param socket_fn: socket文件名称或者共享内存名称
    :param cmd_handler: cmd处理函数，callable类型
    """

    # TODO 使用ZeroMQ进行重新对接
    if ABuEnv.g_is_mac_os:
        from ..UtilBu.ABuMacUtil import socket_bind_recv as do_socket_bind_recv
    else:
        from ..UtilBu.ABuWinUtil import socket_bind_recv as do_socket_bind_recv
    do_socket_bind_recv(socket_fn, cmd_handler)


def socket_send_msg(socket_fn, msg):
    """
    进程间socket或全局共享内通信，发送消息，外层应处理catch_error
    :param socket_fn: socket文件名称或者共享内存名称
    :param msg: 字符串类型需要传递的数据，不需要encode，内部进行encode
    """
    # TODO 使用ZeroMQ进行重新对接
    if ABuEnv.g_is_mac_os:
        from ..UtilBu.ABuMacUtil import socket_send_msg as do_socket_send_msg
    else:
        from ..UtilBu.ABuWinUtil import socket_send_msg as do_socket_send_msg
    do_socket_send_msg(socket_fn, msg)


@catch_error(return_val=0)
def fold_free_size_mb(folder):
    """
    统一平路径可用空间获取，被catch_error装饰，即不应该被提示中断程序，return_val＝0，出错也返回0
    :param folder: 路径或盘符信息
    :return: folder下的可用空间大小
    """
    if ABuEnv.g_is_mac_os:
        from ..UtilBu.ABuMacUtil import fold_free_size_mb as do_fold_free_size_mb
    else:
        from ..UtilBu.ABuWinUtil import fold_free_size_mb as do_fold_free_size_mb
    do_fold_free_size_mb(folder)



from __future__ import absolute_import

from .ABuUmpBase import CachedUmpManager, AbuUmpBase, BuyUmpMixin, SellUmpMixin

from .ABuUmpEdgeBase import AbuUmpEdgeBase
from .ABuUmpEdgeDeg import AbuUmpEdgeDeg, AbuUmpEegeDegExtend
from .ABuUmpEdgeFull import AbuUmpEdgeFull
from .ABuUmpEdgeMul import AbuUmpEdgeMul
from .ABuUmpEdgePrice import AbuUmpEdgePrice
from .ABuUmpEdgeWave import AbuUmpEdgeWave

from .ABuUmpMainBase import AbuUmpMainBase
from .ABuUmpMainDeg import AbuUmpMainDeg, AbuUmpMainDegExtend
from .ABuUmpMainFull import AbuUmpMainFull
from .ABuUmpMainJump import AbuUmpMainJump
from .ABuUmpMainMul import AbuUmpMainMul
from .ABuUmpMainPrice import AbuUmpMainPrice
from .ABuUmpMainWave import AbuUmpMainWave

from . import ABuUmp as ump

__all__ = [
    'CachedUmpManager',
    'AbuUmpBase',
    'BuyUmpMixin',
    'SellUmpMixin',

    'AbuUmpEdgeBase',
    'AbuUmpEdgeDeg',
    'AbuUmpEegeDegExtend',
    'AbuUmpEdgeFull',
    'AbuUmpEdgeMul',
    'AbuUmpEdgePrice',
    'AbuUmpEdgeWave',

    'AbuUmpMainBase',
    'AbuUmpMainDeg',
    'AbuUmpMainDegExtend',
    'AbuUmpMainFull',
    'AbuUmpMainJump',
    'AbuUmpMainMul',
    'AbuUmpMainPrice',
    'AbuUmpMainWave',

    'ump']



# -*- encoding:utf-8 -*-
from __future__ import absolute_import

# noinspection all
from . import ABuSlippageBuyBase as sbb
# noinspection all
from . import ABuSlippageSellBase as ssb
# noinspection all
from . import ABuSlippageBuyMean as sbm



# -*- encoding:utf-8 -*-
"""上层回测图形可视化"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import logging

import ipywidgets as widgets

from ..UtilBu import ABuProgress, ABuFileUtil
from ..WidgetBu.ABuWGBase import WidgetBase, show_msg_func, browser_down_csv_zip
from ..WidgetBu.ABuWGBRunBase import WidgetRunTT
from ..WidgetBu.ABuWGBSymbol import WidgetSymbolChoice
from ..WidgetBu.ABuWGBFBase import BuyFactorWGManager
from ..WidgetBu.ABuWGSFBase import SellFactorWGManager
from ..WidgetBu.ABuWGPSBase import PickStockWGManager

from ..CoreBu.ABu import run_loop_back
from ..CoreBu import ABuEnv
from ..CoreBu.ABuEnv import EDataCacheType, EMarketDataFetchMode, EMarketTargetType
# noinspection PyUnresolvedReferences
from ..CoreBu.ABuFixes import filter
from ..MarketBu.ABuMarket import is_in_sand_box
from ..BetaBu import ABuAtrPosition
from ..AlphaBu import ABuPickTimeExecute
from ..TradeBu.ABuBenchmark import AbuBenchmark
from ..TradeBu.ABuCapital import AbuCapital
from ..MetricsBu.ABuMetricsBase import AbuMetricsBase

__author__ = '阿布'
__weixin__ = 'abu_quant'


# noinspection PyProtectedMember
class WidgetRunLoopBack(WidgetBase):
    """基础界面可以化：初始资金，回测开始，结束周期，参考大盘等"""

    # noinspection PyProtectedMember
    def __init__(self):
        """构建回测需要的各个组件形成tab"""
        self.tt = WidgetRunTT()
        self.sc = WidgetSymbolChoice()
        self.bf = BuyFactorWGManager()
        self.sf = SellFactorWGManager()
        # 卖出策略管理注册买入策略接收改变
        self.sf.register(self.bf)
        self.ps = PickStockWGManager()
        # 选股策略管理注册买入策略接收改变
        self.ps.register(self.bf)

        sub_widget_tab = widgets.Tab()
        sub_widget_tab.children = [self.tt.widget, self.sc.widget, self.bf.widget, self.sf.widget, self.ps.widget]
        for ind, name in enumerate([u'基本', u'股池', u'买策', u'卖策', u'选股']):
            sub_widget_tab.set_title(ind, name)

        self.run_loop_bt = widgets.Button(description=u'开始回测', layout=widgets.Layout(width='98%'),
                                          button_style='danger')
        self.run_loop_bt.on_click(self.run_loop_back)
        self.widget = widgets.VBox([sub_widget_tab, self.run_loop_bt])

    def check_symbol_data_mode(self, choice_symbols):
        """检测是否需要提示下载csv数据或者使用数据下载界面进行操作"""
        if ABuEnv._g_enable_example_env_ipython and choice_symbols is not None:
            # 沙盒模式下 and choice_symbols不是none
            not_in_sb_list = list(filter(lambda symbol: not is_in_sand_box(symbol), choice_symbols))
            if len(not_in_sb_list) > 0:
                logging.info(
                    u'当前数据模式为\'沙盒模式\'无{}数据，'
                    u'请在\'分析设置\'中切换数据模式并确认数据可获取！'
                    u'非沙盒模式建议先用\'数据下载界面操作\'进行数据下载'
                    u'之后设置数据模式为\'开放数据模式\'，联网模式使用\'本地数据模式\''.format(not_in_sb_list))
                browser_down_csv_zip()
                return False

        is_stock_market = \
            ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_FUTURES_CN or \
            ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_US or \
            ABuEnv.g_market_target == EMarketTargetType.E_MARKET_TARGET_HK

        if is_stock_market and not ABuEnv._g_enable_example_env_ipython and choice_symbols is None:
            # 非沙盒模式下要做全股票市场全市场回测
            if ABuEnv.g_data_fetch_mode != EMarketDataFetchMode.E_DATA_FETCH_FORCE_LOCAL:
                logging.info(
                    u'未选择任何回测目标且在非沙盒数据模式下，判定为进行全市场回测'
                    u'为了提高运行效率，请将联网模式修改为\'本地数据模式\'，如需要进行数据更新，'
                    u'请先使用\'数据下载界面操作\'进行数据更新！')
                browser_down_csv_zip()
                return False
            else:
                if ABuEnv.g_data_cache_type == EDataCacheType.E_DATA_CACHE_CSV:
                    # csv模式下
                    if not ABuFileUtil.file_exist(ABuEnv.g_project_kl_df_data_csv):
                        # 股票类型全市场回测，但没有数据
                        logging.info(
                            u'未选择任何回测目标且在非沙盒数据模式下，判定为进行全市场回测'
                            u'为了提高运行效率, 只使用\'本地数据模式\'进行回测，但未发现本地缓存数据，'
                            u'如需要进行数据更新'
                            u'请先使用\'数据下载界面操作\'进行数据更新！')
                        browser_down_csv_zip()
                        return False
                    elif len(os.listdir(ABuEnv.g_project_kl_df_data_csv)) < 100:
                        # 股票类型全市场回测，但数据不足
                        logging.info(
                            u'未选择任何回测目标且在非沙盒数据模式下，判定为进行全市场回测'
                            u'为了提高运行效率, 只使用\'本地数据模式\'进行回测，发现本地缓存数据不足，'
                            u'只有{}支股票历史数据信息'
                            u'如需要进行数据更新'
                            u'请先使用\'数据下载界面操作\'进行数据更新！'.format(
                                len(os.listdir(ABuEnv.g_project_kl_df_data_csv))))
                        browser_down_csv_zip()
                        return False
                elif ABuEnv.g_data_cache_type == EDataCacheType.E_DATA_CACHE_HDF5 \
                        and not ABuFileUtil.file_exist(ABuEnv.g_project_kl_df_data):
                    # hdf5模式下文件不存在
                    logging.info(
                        u'未选择任何回测目标且在非沙盒数据模式下，判定为进行全市场回测'
                        u'为了提高运行效率, 只使用\'本地数据模式\'进行回测'
                        u'hdf5模式下未发现hdf5本地缓存数据，'
                        u'如需要进行数据更新'
                        u'请先使用\'数据下载界面操作\'进行数据更新！')
                    browser_down_csv_zip()
                    return False
        return True

    # noinspection PyUnusedLocal
    def run_loop_back(self, bt):
        """运行回测所对应的button按钮"""
        # 清理之前的输出结果
        ABuProgress.clear_output()

        base_run = self.tt
        # 初始资金
        cash = base_run.cash.value
        n_folds = 2
        start = None
        end = None
        if not base_run.run_years.disabled:
            # 如果使用年回测模式
            n_folds = base_run.run_years.value
        if not base_run.start.disabled:
            # 使用开始回测日期
            start = base_run.start.value
        if not base_run.end.disabled:
            # 使用结束回测日期
            end = base_run.end.value

        choice_symbols = self.sc.choice_symbols.options
        if choice_symbols is not None and len(choice_symbols) == 0:
            # 如果一个symbol都没有设置None， 将使用选择的市场进行全市场回测
            choice_symbols = None

        if not self.check_symbol_data_mode(choice_symbols):
            return

        # 买入策略构成序列
        buy_factors = list(self.bf.factor_dict.values())
        if len(buy_factors) == 0:
            msg = u'没有添加任何一个买入策略！'
            show_msg_func(msg)
            return

        # 卖出策略可以一个也没有
        sell_factors = list(self.sf.factor_dict.values())

        if choice_symbols is not None and len(choice_symbols) == 1:
            # 如果只有1支股票回测，直接使用这个股票做为做为对比基准
            benchmark = AbuBenchmark(choice_symbols[0])
            capital = AbuCapital(cash, benchmark)
            # 如果只有1支股票回测，持仓比例调高
            ABuAtrPosition.g_atr_pos_base = 0.5
            # 就一只股票的情况下也不运行选股策略
            orders_pd, action_pd, _ = ABuPickTimeExecute.do_symbols_with_same_factors(choice_symbols,
                                                                                      benchmark,
                                                                                      buy_factors,
                                                                                      sell_factors,
                                                                                      capital, show=True)
            metrics = AbuMetricsBase(orders_pd, action_pd, capital, benchmark)
        else:
            # 针对选股策略中需要choice_symbols的情况进行选股策略choice_symbols更新
            self.ps.seed_choice_symbol_update(choice_symbols)
            # 多只的情况下使用选股策略
            stock_picks = list(self.ps.factor_dict.values())
            if len(stock_picks) == 0:
                stock_picks = None

            # 多只股票使用run_loop_back
            abu_result_tuple, _ = run_loop_back(cash,
                                                buy_factors,
                                                sell_factors,
                                                stock_picks,
                                                choice_symbols=choice_symbols,
                                                start=start,
                                                end=end,
                                                n_folds=n_folds)
            if abu_result_tuple is None:
                return
            ABuProgress.clear_output()
            metrics = AbuMetricsBase(*abu_result_tuple)
        metrics.fit_metrics()
        metrics.plot_returns_cmp(only_show_returns=True)
